#include <stdlib.h>
#include <string.h>
#ifdef HAVE_IMAGE_MAGICK
# include <stdio.h>
# include <sys/types.h>
# include <wand/magick_wand.h>
#endif
#include <gtk/gtk.h>

#include "guiutils.h"
#include "csd.h"

#include "convparm.h"

#include "images/icon_convparm_fmt_20x20.xpm"
#include "images/icon_convparm_compression_20x20.xpm"
#include "images/icon_convparm_colorbg_20x20.xpm"
#include "images/icon_convparm_gamma_20x20.xpm"
#include "images/icon_convparm_mirror_20x20.xpm"
#include "images/icon_convparm_crop_20x20.xpm"
#include "images/icon_convparm_rotate_20x20.xpm"
#include "images/icon_convparm_size_20x20.xpm"
#include "images/icon_convparm_frame_20x20.xpm"
#include "images/icon_convparm_text_20x20.xpm"


typedef struct _convparm_fmt_struct	convparm_fmt_struct;
typedef struct _convparm_compression_struct convparm_compression_struct;
typedef struct _convparm_colorbg_struct	convparm_colorbg_struct;
typedef struct _convparm_gamma_struct	convparm_gamma_struct;
typedef struct _convparm_mirror_struct	convparm_mirror_struct;
typedef struct _convparm_crop_struct	convparm_crop_struct;
typedef struct _convparm_rotate_struct	convparm_rotate_struct;
typedef struct _convparm_size_struct	convparm_size_struct;
typedef struct _convparm_text_struct	convparm_text_struct;
typedef struct _convparm_frame_struct	convparm_frame_struct;


/*
 *	Conversion Parameter Format:
 */
struct _convparm_fmt_struct {
	convparm_struct	convparm;
	GtkWidget	*fmt_combo;
};
#define CONVPARM_FMT(p)		((convparm_fmt_struct *)(p))

/*
 *	Conversion Parameter Compression:
 */
struct _convparm_compression_struct {
	convparm_struct convparm;
	GtkWidget       *scale,
			*entry;
};
#define CONVPARM_COMPRESSION(p)	((convparm_compression_struct *)(p))


/*
 *	Conversion Parameter Background Color:
 */
struct _convparm_colorbg_struct {
	convparm_struct	convparm;
	GtkWidget	*btn;
	GdkColor	c;
};
#define CONVPARM_COLORBG(p)	((convparm_colorbg_struct *)(p))

/*
 *	Conversion Parameter Gamma:
 */
struct _convparm_gamma_struct {
	convparm_struct convparm;
	GtkWidget	*bar_pm,
			*scale,
			*entry;
};
#define CONVPARM_GAMMA(p)	((convparm_gamma_struct *)(p))

/*
 *      Mirror:
 */
struct _convparm_mirror_struct {
	convparm_struct	convparm;
	GtkWidget	*horizontal_radio,
			*vertical_radio;
};
#define CONVPARM_MIRROR(p)	((convparm_mirror_struct *)(p))

/*
 *	Conversion Parameter Crop:
 */
struct _convparm_crop_struct {
	convparm_struct	convparm;
	GtkWidget	*x_spin,
			*y_spin,
			*width_spin,
			*height_spin;
	GdkRectangle	rect;
};
#define CONVPARM_CROP(p)	((convparm_crop_struct *)(p))

/*
 *      Conversion Parameter Rotate:
 */
struct _convparm_rotate_struct {
	convparm_struct	convparm;
	GtkWidget	*degrees_spin;
};
#define CONVPARM_ROTATE(p)	((convparm_rotate_struct *)(p))

/*
 *	Conversion Parameter Size:
 */
struct _convparm_size_struct {
	convparm_struct	convparm;
	GtkWidget	*width_spin,
			*height_spin,
			*maintain_aspect_tb,
			*smaller_tb;
	convparm_sizeinfo_struct	sizeinfo;
};
#define CONVPARM_SIZE(p)	((convparm_size_struct *)(p))

/*
 *      Conversion Parameter Text:
 */
struct _convparm_text_struct {
	convparm_struct	convparm;
	GtkWidget	*prev_btn,
			*entry,
			*next_btn,
			*add_btn,
			*edit_btn,
			*remove_btn;
	gint		sel_textline;
	GList		*textlines;
};
#define CONVPARM_TEXT(p)	((convparm_text_struct *)(p))

/*
 *	Conversion Parameter Frame:
 */
struct _convparm_frame_struct {
	convparm_struct	convparm;
	GtkWidget       *width_spin,
			*height_spin,
			*inner_bevel_spin,
			*outer_bevel_spin,
			*color_btn;
	convparm_frameinfo_struct	frameinfo;
};
#define CONVPARM_FRAME(p)	((convparm_frame_struct *)(p))


/* Callbacks */
static void ConvParmToggleEnableCB(GtkWidget *widget, gpointer data);
static void ConvParmToggleSensitiveCB(GtkWidget *widget, gpointer data);

/* Basics */
static convparm_struct *ConvParmNew(
	convparm_type type,
	gint size,		/* Of the struct in bytes */
	guint8 **icon_data,
	const gchar *name,
	GtkWidget *parent,
	void (*destroy_func)(gpointer)
);
void ConvParmDelete(convparm_struct *p);
void ConvParmShow(convparm_struct *p);
void ConvParmHide(convparm_struct *p);
void ConvParmEnable(convparm_struct *p);
void ConvParmDisable(convparm_struct *p);
convparm_struct *ConvParmListMatchType(
	GList *convparm_list, convparm_type type
);

/* Format */
convparm_struct *ConvParmFmtNew(GtkWidget *parent);
static void ConvParmFmtDelete(gpointer data);
void ConvParmFmtSet(convparm_struct *p, const gchar *fmt);
const gchar *ConvParmFmtGet(convparm_struct *p);
gboolean ConvParmFmtCompare(const gchar *s1, const gchar *s2);

/* Compression */
convparm_struct *ConvParmCompressionNew(GtkWidget *parent);
static void ConvParmCompressionDelete(gpointer data);
void ConvParmCompressionSet(convparm_struct *p, gfloat compression);
gfloat ConvParmCompressionGet(convparm_struct *p);
static void ConvParmCompressionValueChangedCB(
	GtkAdjustment *adj, gpointer data
);

/* Background Color */
convparm_struct *ConvParmColorBGNew(GtkWidget *parent);
static void ConvParmColorBGDelete(gpointer data);
void ConvParmColorBGSet(convparm_struct *p, const GdkColor *c);
const GdkColor *ConvParmColorBGGet(convparm_struct *p);

/* Gamma */
convparm_struct *ConvParmGammaNew(GtkWidget *parent);
static void ConvParmGammaDelete(gpointer data);
void ConvParmGammaSet(convparm_struct *p, gfloat g);
gfloat ConvParmGammaGet(convparm_struct *p);
static void ConvParmGammaAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void ConvParmGammaEntryChangedCB(GtkEntry *entry, gpointer data);

/* Mirror */
convparm_struct *ConvParmMirrorNew(GtkWidget *parent);
static void ConvParmMirrorDelete(gpointer data);
void ConvParmMirrorSet(convparm_struct *p, GtkOrientation orientation);
GtkOrientation ConvParmMirrorGet(convparm_struct *p);

/* Crop */
convparm_struct *ConvParmCropNew(GtkWidget *parent);
static void ConvParmCropDelete(gpointer data);
void ConvParmCropSet(convparm_struct *p, const GdkRectangle *rect);
const GdkRectangle *ConvParmCropGet(convparm_struct *p);

/* Rotate */
convparm_struct *ConvParmRotateNew(GtkWidget *parent);
static void ConvParmRotateDelete(gpointer data);
void ConvParmRotateSet(convparm_struct *p, gfloat deg);
gfloat ConvParmRotateGet(convparm_struct *p);

/* Size */
convparm_struct *ConvParmSizeNew(GtkWidget *parent);
static void ConvParmSizeDelete(gpointer data);
void ConvParmSizeSet(
	convparm_struct *p, const convparm_sizeinfo_struct *s
);
const convparm_sizeinfo_struct *ConvParmSizeGet(convparm_struct *p);

/* Text */
convparm_struct *ConvParmTextNew(GtkWidget *parent);
static void ConvParmTextDelete(gpointer data);
void ConvParmTextSet(convparm_struct *p, GList *textlines);
GList *ConvParmTextGet(convparm_struct *p);
void ConvParmTextLineDelete(convparm_textline_struct *l);

/* Frame */
convparm_struct *ConvParmFrameNew(GtkWidget *parent);
static void ConvParmFrameDelete(gpointer data);
void ConvParmFrameSet(
	convparm_struct *p, const convparm_frameinfo_struct *f
);
const convparm_frameinfo_struct *ConvParmFrameGet(convparm_struct *p);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)

#ifdef HAVE_IMAGE_MAGICK
#define MAGICK_FREE(p)		(((p) != NULL) ? \
 MagickRelinquishMemory(p) : NULL)
#endif


/*
 *	GtkToggleButton "toggled" signal callback.
 *
 *	Updates the enabled or disabled state of the conversion
 *	parameter based on the state of the GtkToggleButton.
 */
static void ConvParmToggleEnableCB(GtkWidget *widget, gpointer data)
{
	convparm_struct *p = CONVPARM(data);
	if((widget == NULL) || (p == NULL))
	    return;

	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(widget))
	    p->flags |= CONVPARM_ENABLED;
	else
	    p->flags &= ~CONVPARM_ENABLED;
}

/*
 *	GtkToggleButton "toggled" signal callback.
 *
 *	Sets the widget specified by data to be sensitive or insensitive
 *	based on the toggled state of the GtkToggleButton.
 */
static void ConvParmToggleSensitiveCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w = (GtkWidget *)data;
	if(w == NULL)
	    return;

	gtk_widget_set_sensitive(
	    w, GTK_TOGGLE_BUTTON_GET_ACTIVE(widget)
	);
}


/*
 *	Creates a new Conversion Parameter.
 */
static convparm_struct *ConvParmNew(
	convparm_type type,
	gint size,			/* Of the struct in bytes */
	guint8 **icon_data,
	const gchar *name,
	GtkWidget *parent,
	void (*destroy_func)(gpointer)
)
{
	convparm_struct *p = CONVPARM(g_malloc0(size));
	if(p == NULL)
	    return(NULL);

	p->type = type;
	p->flags = CONVPARM_ENABLED;
	p->freeze_count = 0;
	p->name = STRDUP(name);
	p->destroy_func = destroy_func;
	p->destroy_data = p;

	if(parent != NULL)
	{
	    const gint	border_major = 5,
			border_minor = 2;
	    GtkWidget *w, *parent2, *parent3;

	    p->toplevel = w = gtk_hbox_new(FALSE, border_major);
	    if(GTK_IS_BOX(parent))
		gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    else if(GTK_IS_CONTAINER(parent))
		gtk_container_add(GTK_CONTAINER(parent), w);
	    parent = w;

	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
            parent2 = w;

	    p->tb = w = gtk_check_button_new();
	    GTK_TOGGLE_BUTTON_SET_ACTIVE(w, CONVPARM_IS_ENABLED(p));
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(w), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleEnableCB), p
	    );
	    gtk_widget_show(w);
	    parent3 = w;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_show(w);
	    parent3 = w;

	    if(icon_data != NULL)
	    {
		p->icon_pm = w = gtk_pixmap_new_from_xpm_d(
		    NULL, NULL, icon_data
		);
		gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
		gtk_widget_show(w);
	    }

	    p->label = w = gtk_label_new(
		(name != NULL) ? name : "Untitled"
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	}

	return(p);
}

void ConvParmDelete(convparm_struct *p)
{
	if(p == NULL)
	    return;

	if(p->destroy_func != NULL)
	    p->destroy_func(p->destroy_data);
	GTK_WIDGET_DESTROY(p->icon_pm);
	GTK_WIDGET_DESTROY(p->label);
	GTK_WIDGET_DESTROY(p->tb);
	GTK_WIDGET_DESTROY(p->toplevel);
	g_free(p->name);
	g_free(p);
}

void ConvParmShow(convparm_struct *p)
{
	GtkWidget *w = CONVPARM_TOPLEVEL(p);
	if(w == NULL)
	    return;

	gtk_widget_show(w);
}

void ConvParmHide(convparm_struct *p)
{
	GtkWidget *w = CONVPARM_TOPLEVEL(p);
	if(w == NULL)
	    return;

	gtk_widget_hide(w);
}

void ConvParmEnable(convparm_struct *p)
{
	GTK_TOGGLE_BUTTON_SET_ACTIVE(CONVPARM_TB(p), TRUE);
}

void ConvParmDisable(convparm_struct *p)
{
	GTK_TOGGLE_BUTTON_SET_ACTIVE(CONVPARM_TB(p), FALSE);
}

convparm_struct *ConvParmListMatchType(
	GList *convparm_list, convparm_type type
)
{
	convparm_struct *p;

	for(;
	    convparm_list != NULL;
	    convparm_list = g_list_next(convparm_list)
	)
	{
	    p = CONVPARM(convparm_list->data);
	    if(p == NULL)
		continue;
	    if(p->type == type)
		return(p);
	}

	return(NULL);
}


/* Format */
convparm_struct *ConvParmFmtNew(GtkWidget *parent)
{
	convparm_fmt_struct *pv = CONVPARM_FMT(ConvParmNew(
	    CONVPARM_TYPE_FMT,
	    sizeof(convparm_fmt_struct),
	    (guint8 **)icon_convparm_fmt_20x20_xpm,
	    "Format",
	    parent,
	    ConvParmFmtDelete
	));
	if(pv == NULL)
	    return(NULL);

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint border_minor = 2;
	    GtkWidget *w, *parent2;
	    GtkCombo *combo;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w 
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    pv->fmt_combo = w = gtk_combo_new();
	    combo = GTK_COMBO(w);
	    gtk_widget_set_usize(w, 150, -1);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
#ifdef HAVE_IMAGE_MAGICK
	    if(TRUE)
	    {
		gint i;
		GList *glist = NULL;
		unsigned long n;
		char **strv = MagickQueryFormats("*", &n);
		glist = g_list_append(glist, STRDUP("*"));
		for(i = 0; i < (gint)n; i++)
		{
		    glist = g_list_append(glist, STRDUP(strv[i]));
		    MAGICK_FREE(strv[i]);
		}
		MAGICK_FREE(strv);
		gtk_combo_set_popdown_strings(combo, glist);
		g_list_foreach(glist, (GFunc)g_free, NULL);
		g_list_free(glist);
	    }
#endif
	    gtk_widget_show(w);

	    w = combo->entry;
	    GUIEditableEndowPopupMenu(w, 0);
	}

	return(CONVPARM(pv));
}

static void ConvParmFmtDelete(gpointer data)
{
	convparm_fmt_struct *pv = CONVPARM_FMT(data);
	if(!CONVPARM_IS_FMT(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->fmt_combo);
}

void ConvParmFmtSet(convparm_struct *p, const gchar *fmt)
{
	GtkCombo *combo;
	convparm_fmt_struct *pv = CONVPARM_FMT(p);
	if(!CONVPARM_IS_FMT(pv) || (fmt == NULL))
	    return;

	combo = (GtkCombo *)pv->fmt_combo;
	if(combo != NULL)
	{
	    GtkEntry *entry = GTK_ENTRY(combo->entry);
	    const gchar *old_fmt = gtk_entry_get_text(entry);
	    if(old_fmt != fmt)
		gtk_entry_set_text(entry, fmt);
	}
}

const gchar *ConvParmFmtGet(convparm_struct *p)
{
	GtkCombo *combo;
	convparm_fmt_struct *pv = CONVPARM_FMT(p);
	if(!CONVPARM_IS_FMT(pv)) 
	    return(NULL);

	combo = (GtkCombo *)pv->fmt_combo;
	if(combo == NULL)
	    return(NULL);

	return(gtk_entry_get_text(GTK_ENTRY(combo->entry)));
}

gboolean ConvParmFmtCompare(const gchar *s1, const gchar *s2)
{
	const gchar *ext_ptr;

	if((s1 == NULL) || (s2 == NULL))
	    return(FALSE);

	ext_ptr = strrchr(s1, '.');
	if(ext_ptr != NULL)
	    s1 = ext_ptr;
	while((*s1 == '.') || (*s1 == ' ') || (*s1 == '\t'))
	    s1++;

	ext_ptr = strrchr(s2, '.');
	if(ext_ptr != NULL)
	    s2 = ext_ptr;
	while((*s2 == '.') || (*s2 == ' ') || (*s2 == '\t'))
	    s2++;

	if(!g_strcasecmp(s1, "*") || !g_strcasecmp(s2, "*"))
	    return(TRUE);

	if(!g_strcasecmp(s1, s2))
	    return(TRUE);

	return(FALSE);
}


/* Compression */
convparm_struct *ConvParmCompressionNew(GtkWidget *parent)
{
	convparm_compression_struct *pv = CONVPARM_COMPRESSION(ConvParmNew(
	    CONVPARM_TYPE_COMPRESSION,
	    sizeof(convparm_compression_struct),
	    (guint8 **)icon_convparm_compression_20x20_xpm,
	    "Compression",
	    parent,
	    ConvParmCompressionDelete  
	));
	if(pv == NULL)
	    return(NULL);

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint	border_minor = 2;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, 1.0f, 0.1f, 0.1f, 0.0f
	    );
	    pv->scale = w = gtk_hscale_new(adj);
	    gtk_widget_set_usize(w, 150, -1);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(ConvParmCompressionValueChangedCB), pv
	    );
	    GUISetWidgetTip(
		w,
"Set the compression for images that are compressable.\
 Higher values increase the amount of compression and processing time.\
 For image formats that are lossy, higher values will decrease the quality.\
 For JPEG images the default value is 20%.\
 For PNG images the default value is 67%."
	    );
	    gtk_widget_show(w);

	    pv->entry = w = gtk_entry_new();
	    gtk_widget_set_usize(w, 125, -1);
	    gtk_entry_set_editable(GTK_ENTRY(w), FALSE);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, GUI_EDITABLE_POPUP_MENU_READ_ONLY);
	    gtk_widget_show(w);
	}

	return(CONVPARM(pv));
}

static void ConvParmCompressionDelete(gpointer data)
{
	convparm_compression_struct *pv = CONVPARM_COMPRESSION(data);
	if(!CONVPARM_IS_COMPRESSION(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->scale);
	GTK_WIDGET_DESTROY(pv->entry);
}

void ConvParmCompressionSet(convparm_struct *p, gfloat compression)
{
	GtkAdjustment *adj;
	GtkWidget *w;
	convparm_compression_struct *pv = CONVPARM_COMPRESSION(p);
	if(!CONVPARM_IS_COMPRESSION(pv))
	    return;

	w = pv->scale;
	adj = (w != NULL) ? GTK_RANGE(w)->adjustment : NULL;
	if(adj != NULL)
	    gtk_adjustment_set_value(adj, compression);
}

gfloat ConvParmCompressionGet(convparm_struct *p)
{
	GtkAdjustment *adj;
	GtkWidget *w;
	convparm_compression_struct *pv = CONVPARM_COMPRESSION(p);
	if(!CONVPARM_IS_COMPRESSION(pv))
	    return(0.0f);

	w = pv->scale;
	adj = (w != NULL) ? GTK_RANGE(w)->adjustment : NULL;
	return((adj != NULL) ? adj->value : 0.0f);
}

static void ConvParmCompressionValueChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	GtkWidget *w;
	convparm_compression_struct *pv = CONVPARM_COMPRESSION(data);
	if(!CONVPARM_IS_COMPRESSION(pv))
	    return;

	w = pv->entry;
	if(w != NULL)
	{
	    const gfloat compression = ConvParmCompressionGet(
		CONVPARM(pv)
	    );
	    gchar *s;
	    if(compression >= 1.0f)
		s = g_strdup_printf(
		    "%i%% (Maximum)",
		    (gint)(compression * 100.0f)
		);
	    else if(compression >= 0.66f)
		s = g_strdup_printf(
		    "%i%% (Optimul)",
		    (gint)(compression * 100.0f)
		);
	    else if(compression >= 0.33f)
		s = g_strdup_printf(
		    "%i%% (Medium)",
		    (gint)(compression * 100.0f)
		);
	    else if(compression > 0.0f)
		s = g_strdup_printf(
		    "%i%% (Low)",
		    (gint)(compression * 100.0f)
		);
	    else
		s = STRDUP("No Compression");
	    gtk_entry_set_text(GTK_ENTRY(w), s);
	    g_free(s);
	}
}


/* Background Color */
convparm_struct *ConvParmColorBGNew(GtkWidget *parent)
{
	convparm_colorbg_struct *pv = CONVPARM_COLORBG(ConvParmNew(
	    CONVPARM_TYPE_COLORBG, 
	    sizeof(convparm_colorbg_struct),
	    (guint8 **)icon_convparm_colorbg_20x20_xpm,
	    "Background Color",
	    parent,
	    ConvParmColorBGDelete
	));
	if(pv == NULL)
	    return(NULL);

	memset(&pv->c, 0x00, sizeof(GdkColor));

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint border_minor = 2;
	    GtkWidget *w, *parent2;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    pv->btn = w = CSDColorButtonNewSimple(NULL, -1);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	}

	return(CONVPARM(pv));
}

static void ConvParmColorBGDelete(gpointer data)
{
	convparm_colorbg_struct *pv = CONVPARM_COLORBG(data);
	if(!CONVPARM_IS_COLORBG(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->btn);
}

void ConvParmColorBGSet(convparm_struct *p, const GdkColor *c)
{
	csd_color_struct csd_c;

	convparm_colorbg_struct *pv = CONVPARM_COLORBG(p);
	if(!CONVPARM_IS_COLORBG(pv) || (c == NULL))
	    return;

	memcpy(&pv->c, c, sizeof(GdkColor));
	csd_c.a = 1.0f;
	csd_c.r = (gfloat)c->red / (gfloat)((gushort)-1);
	csd_c.g = (gfloat)c->green / (gfloat)((gushort)-1);
	csd_c.b = (gfloat)c->blue / (gfloat)((gushort)-1);
	CSDColorButtonSetColor(pv->btn, &csd_c);
}

const GdkColor *ConvParmColorBGGet(convparm_struct *p)
{
	GdkColor *c;
	csd_color_struct csd_c;
	convparm_colorbg_struct *pv = CONVPARM_COLORBG(p);
	if(!CONVPARM_IS_COLORBG(pv))
	    return(NULL);
	c = &pv->c;
	CSDColorButtonGetColor(pv->btn, &csd_c);
	GDK_COLOR_SET_COEFF(c, csd_c.r, csd_c.g, csd_c.b);
	return(c);
}


/* Gamma */
convparm_struct *ConvParmGammaNew(GtkWidget *parent)
{
	convparm_gamma_struct *pv = CONVPARM_GAMMA(ConvParmNew(
	    CONVPARM_TYPE_GAMMA,
	    sizeof(convparm_gamma_struct),
	    (guint8 **)icon_convparm_gamma_20x20_xpm,
	    "Gamma",
	    parent,
	    ConvParmGammaDelete
	));
	if(pv == NULL)
	    return(NULL);

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint	border_minor = 2,
			bar_width = 200,
			bar_height = 10,
			scale_handle_width = (2 * 16);
	    gint i, width, height;
	    GdkColor c_buf, *c = &c_buf;
	    GdkColormap *colormap = gdk_colormap_get_system();
	    GdkPixmap *pixmap;
	    GdkGC *gc = GDK_GC_NEW(NULL);
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3, *parent4;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    /* GtkVbox for the bar GtkPixmap and the GtkScale */
	    w = gtk_vbox_new(FALSE, 0);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* GtkAlignment to align the bar GtkPixmap */
	    width = bar_width + scale_handle_width + (2 * 2);
	    height = bar_height;
	    w = gtk_alignment_new(0.5f, 0.5f, 0.0f, 0.0f);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, width, height);
	    gtk_widget_show(w);
	    parent4 = w;

	    /* Bar GtkPixmap */
	    width = bar_width;
	    height = bar_height;
	    pixmap = GDK_PIXMAP_NEW(width, height);
	    for(i = 0; i < width; i++)
	    {
		const gfloat g = (gfloat)(i + 1) / (gfloat)width;
		GDK_COLOR_SET_COEFF(c, g, g, g);
		GDK_COLORMAP_ALLOC_COLOR(colormap, c);
		gdk_gc_set_foreground(gc, c);
		gdk_draw_line(
		    (GdkDrawable *)pixmap, gc,
		    i, 0, i, height
		);
		GDK_COLORMAP_FREE_COLOR(colormap, c);
	    }
	    pv->bar_pm = w = gtk_pixmap_new(pixmap, NULL);
	    GDK_PIXMAP_UNREF(pixmap);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    gtk_widget_show(w);

	    /* GtkScale */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, -1.0f, 1.0f, 0.1f, 0.1f, 0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(ConvParmGammaAdjustmentValueChangedCB), pv
	    );
	    pv->scale = w = gtk_hscale_new(adj);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	    gtk_widget_show(w);

	    /* GtkEntry */
	    pv->entry = w = gtk_entry_new();
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_set_usize(w, 50, -1);
	    gtk_entry_set_text(GTK_ENTRY(w), "1.0");
	    gtk_signal_connect(
		GTK_OBJECT(w), "changed",
		GTK_SIGNAL_FUNC(ConvParmGammaEntryChangedCB), pv
	    );
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);


	    GDK_GC_UNREF(gc);
	}

	return(CONVPARM(pv));
}

static void ConvParmGammaDelete(gpointer data)
{
	convparm_gamma_struct *pv = CONVPARM_GAMMA(data);
	if(!CONVPARM_IS_GAMMA(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->bar_pm);
	GTK_WIDGET_DESTROY(pv->scale);
}

void ConvParmGammaSet(convparm_struct *p, gfloat g)
{
	convparm_gamma_struct *pv = CONVPARM_GAMMA(p);
	if(!CONVPARM_IS_GAMMA(pv))
	    return;

	if(g < 1.0f)
	    g = g - 1.0f;
	else
	    g = (g - 1.0f) / 255.0f;

	gtk_adjustment_set_value(
	    gtk_range_get_adjustment(GTK_RANGE(pv->scale)), g
	);
}

gfloat ConvParmGammaGet(convparm_struct *p)
{
	gfloat g;
	GtkAdjustment *adj;
	convparm_gamma_struct *pv = CONVPARM_GAMMA(p);
	if(!CONVPARM_IS_GAMMA(pv))
	    return(0.0f);

	adj = gtk_range_get_adjustment(GTK_RANGE(pv->scale));
	g = (adj != NULL) ? adj->value : 0.0f;
	if(g < 0.0f)
	    g = g + 1.0f;
	else
	    g = (g * 255.0f) + 1.0f;

	return(g);
}

static void ConvParmGammaAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	gfloat g;
	gchar *s;
	convparm_gamma_struct *pv = CONVPARM_GAMMA(data);
	if((adj == NULL) || (pv == NULL))
	    return;

	if(CONVPARM_IS_FROZEN(pv))
	    return;

	CONVPARM_FREEZE(pv);

	g = adj->value;
	if(g < 0.0f)
	    s = g_strdup_printf(
		"%.1f",
		g + 1.0f
	    );
	else
	    s = g_strdup_printf(
		"%.1f",
		(g * 255.0f) + 1.0f
	    );
	gtk_entry_set_text(GTK_ENTRY(pv->entry), s);
	g_free(s);	

	CONVPARM_THAW(pv);
}

static void ConvParmGammaEntryChangedCB(GtkEntry *entry, gpointer data)
{
	convparm_gamma_struct *pv = CONVPARM_GAMMA(data);
	if((entry == NULL) || (pv == NULL))
	    return;

	if(CONVPARM_IS_FROZEN(pv))
	    return;

	CONVPARM_FREEZE(pv);

	ConvParmGammaSet(
	    CONVPARM(pv),
	    ATOF(gtk_entry_get_text(entry))
	);

	CONVPARM_THAW(pv);
}


/* Mirror */
convparm_struct *ConvParmMirrorNew(GtkWidget *parent)
{
	convparm_mirror_struct *pv = CONVPARM_MIRROR(ConvParmNew(
	    CONVPARM_TYPE_MIRROR,
	    sizeof(convparm_mirror_struct),
	    (guint8 **)icon_convparm_mirror_20x20_xpm,
	    "Mirror",
	    parent,
	    ConvParmMirrorDelete
	));
	if(pv == NULL)
	    return(NULL);

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{                 
	    const gint border_minor = 2;
	    GSList *gslist;
	    GtkWidget *w, *parent2;

	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    gslist = NULL;
	    pv->horizontal_radio = w = gtk_radio_button_new_with_label(
		gslist,
		"Horizontal"
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	    gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	    pv->vertical_radio = w = gtk_radio_button_new_with_label(
		gslist,
		"Vertical"
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);

	}

	return(CONVPARM(pv));
}

static void ConvParmMirrorDelete(gpointer data)
{
	convparm_mirror_struct *pv = CONVPARM_MIRROR(data);
	if(!CONVPARM_IS_MIRROR(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->horizontal_radio);
	GTK_WIDGET_DESTROY(pv->vertical_radio);
}

void ConvParmMirrorSet(convparm_struct *p, GtkOrientation orientation)
{
	convparm_mirror_struct *pv = CONVPARM_MIRROR(p);
	if(!CONVPARM_IS_MIRROR(pv))
	    return;

	switch(orientation)
	{
	  case GTK_ORIENTATION_VERTICAL:
	    GTK_TOGGLE_BUTTON_SET_ACTIVE(pv->vertical_radio, TRUE);
	    break;
	  case GTK_ORIENTATION_HORIZONTAL:
	    GTK_TOGGLE_BUTTON_SET_ACTIVE(pv->horizontal_radio, TRUE);
	    break;
	}
}

GtkOrientation ConvParmMirrorGet(convparm_struct *p)
{
	convparm_mirror_struct *pv = CONVPARM_MIRROR(p);
	if(!CONVPARM_IS_MIRROR(pv))
	    return(GTK_ORIENTATION_HORIZONTAL);

	return(
	    GTK_TOGGLE_BUTTON_GET_ACTIVE(pv->vertical_radio) ?
		GTK_ORIENTATION_VERTICAL : GTK_ORIENTATION_HORIZONTAL
	);
}

/* Crop */
convparm_struct *ConvParmCropNew(GtkWidget *parent)
{
	convparm_crop_struct *pv = CONVPARM_CROP(ConvParmNew(
	    CONVPARM_TYPE_CROP,
	    sizeof(convparm_crop_struct),
	    (guint8 **)icon_convparm_crop_20x20_xpm,
	    "Crop",
	    parent,
	    ConvParmCropDelete
	));
	if(pv == NULL)
	    return(NULL);

	memset(&pv->rect, 0x00, sizeof(GdkRectangle));

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint  border_minor = 2,
			border_major = 5;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3;

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    /* X */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("X:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, (gfloat)((guint16)-1) / -2.0f,
		(gfloat)((guint16)-1) / 2.0f, 1.0f, 10.0f, 0.0f
	    );
	    pv->x_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Y */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("Y:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, (gfloat)((guint16)-1) / -2.0f,
		(gfloat)((guint16)-1) / 2.0f, 1.0f, 10.0f, 0.0f
	    );
	    pv->y_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Width */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("W:"); 
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 10.0f, 0.0f
	    );
	    pv->width_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Height */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("H:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 10.0f, 0.0f
	    );
	    pv->height_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);
	}

	return(CONVPARM(pv));
}

static void ConvParmCropDelete(gpointer data)
{
	convparm_crop_struct *pv = CONVPARM_CROP(data);
	if(!CONVPARM_IS_CROP(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->x_spin);
	GTK_WIDGET_DESTROY(pv->y_spin);
	GTK_WIDGET_DESTROY(pv->width_spin);
	GTK_WIDGET_DESTROY(pv->height_spin);
}

void ConvParmCropSet(convparm_struct *p, const GdkRectangle *rect)
{
	convparm_crop_struct *pv = CONVPARM_CROP(p);
	if(!CONVPARM_IS_CROP(pv) || (rect == NULL))
	    return;

	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->x_spin), (gfloat)rect->x
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->y_spin), (gfloat)rect->y
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->width_spin), (gfloat)rect->width
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->height_spin), (gfloat)rect->height
	);
}

const GdkRectangle *ConvParmCropGet(convparm_struct *p)
{
	GdkRectangle *rect;
	convparm_crop_struct *pv = CONVPARM_CROP(p);
	if(!CONVPARM_IS_CROP(pv))
	    return(NULL);

	rect = &pv->rect;

	rect->x = (gint16)gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->x_spin)
	);
	rect->y = (gint16)gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->y_spin)
	);
	rect->width = (guint16)gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->width_spin)
	);
	rect->height = (guint16)gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->height_spin)
	);

	return(rect);
}


/* Rotate */
convparm_struct *ConvParmRotateNew(GtkWidget *parent)
{
	convparm_rotate_struct *pv = CONVPARM_ROTATE(ConvParmNew(
	    CONVPARM_TYPE_ROTATE,
	    sizeof(convparm_rotate_struct),
	    (guint8 **)icon_convparm_rotate_20x20_xpm,
	    "Rotate",
	    parent,
	    ConvParmRotateDelete
	));
	if(pv == NULL)
	    return(NULL);

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint  border_minor = 2,
			border_major = 5;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3;

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;

	    /* Degrees Clockwise */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, -360.0f, 360.0f, 1.0f, 10.0f, 0.0f
	    );
	    pv->degrees_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);
	    /* GtkLabel */
	    w = gtk_label_new("Degrees Clockwise");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	}
	 
	return(CONVPARM(pv));
}

static void ConvParmRotateDelete(gpointer data)
{
	convparm_rotate_struct *pv = CONVPARM_ROTATE(data);
	if(!CONVPARM_IS_ROTATE(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->degrees_spin);
}

void ConvParmRotateSet(convparm_struct *p, gfloat deg)
{
	convparm_rotate_struct *pv = CONVPARM_ROTATE(p);
	if(!CONVPARM_IS_ROTATE(pv))
	    return;

	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->degrees_spin), deg
	);
}

gfloat ConvParmRotateGet(convparm_struct *p)
{
	convparm_rotate_struct *pv = CONVPARM_ROTATE(p);
	if(!CONVPARM_IS_ROTATE(pv))
	    return(0.0f);

	return(gtk_spin_button_get_value_as_float(
	    GTK_SPIN_BUTTON(pv->degrees_spin)
	));
}


/* Size */
convparm_struct *ConvParmSizeNew(GtkWidget *parent)
{
	convparm_size_struct *pv = CONVPARM_SIZE(ConvParmNew(
	    CONVPARM_TYPE_SIZE,
	    sizeof(convparm_size_struct),
	    (guint8 **)icon_convparm_size_20x20_xpm,
	    "Size",
	    parent,
	    ConvParmSizeDelete
	));
	if(pv == NULL)
	    return(NULL);

	memset(&pv->sizeinfo, 0x00, sizeof(convparm_sizeinfo_struct));

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint	border_minor = 2,
			border_major = 5;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3;

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",  
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;
 
	    /* Width */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("W:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 10.0f, 0.0f
	    );
	    pv->width_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Height */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("H:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 10.0f, 0.0f
	    );
	    pv->height_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Maintain Aspect */
	    pv->maintain_aspect_tb = w =
		gtk_check_button_new_with_label("Maintain Aspect");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Check this to maintain the image's original aspect ratio"
	    );
	    gtk_widget_show(w);

	    /* Smaller */
	    pv->smaller_tb = w =
		gtk_check_button_new_with_label("Smaller");
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Check this to use the smaller of the two dimensions when maintaining\
 aspect ratio, otherwise the larger of the two dimensions will be used"
	    );
	    gtk_widget_show(w);
	}

	return(CONVPARM(pv));
}

static void ConvParmSizeDelete(gpointer data)
{
	convparm_size_struct *pv = CONVPARM_SIZE(data);
	if(!CONVPARM_IS_SIZE(pv))
	    return;
		   
	GTK_WIDGET_DESTROY(pv->width_spin);
	GTK_WIDGET_DESTROY(pv->height_spin);
	GTK_WIDGET_DESTROY(pv->maintain_aspect_tb);
	GTK_WIDGET_DESTROY(pv->smaller_tb);
}

void ConvParmSizeSet(
	convparm_struct *p, const convparm_sizeinfo_struct *s
)
{
	convparm_size_struct *pv = CONVPARM_SIZE(p);
	if(!CONVPARM_IS_SIZE(pv) || (s == NULL))
	    return;

	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->width_spin), (gfloat)s->width
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->height_spin), (gfloat)s->height
	);
	GTK_TOGGLE_BUTTON_SET_ACTIVE(
	    pv->maintain_aspect_tb, s->maintain_aspect
	);
	GTK_TOGGLE_BUTTON_SET_ACTIVE(
	    pv->smaller_tb, s->smaller
	);
}

const convparm_sizeinfo_struct *ConvParmSizeGet(convparm_struct *p)
{
	convparm_sizeinfo_struct *s;
	convparm_size_struct *pv = CONVPARM_SIZE(p);
	if(!CONVPARM_IS_SIZE(pv))
	    return(NULL);

	s = &pv->sizeinfo;

	s->width = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->width_spin)
	);
	s->height = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->height_spin)
	);
	s->maintain_aspect = GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    pv->maintain_aspect_tb
	);
	s->smaller = GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    pv->smaller_tb
	);

	return(s);
}


/* Text */
convparm_struct *ConvParmTextNew(GtkWidget *parent)
{
	convparm_text_struct *pv = CONVPARM_TEXT(ConvParmNew(
	    CONVPARM_TYPE_TEXT,
	    sizeof(convparm_text_struct),
	    (guint8 **)icon_convparm_text_20x20_xpm,
	    "Text",
	    parent,
	    ConvParmTextDelete
	));
	if(pv == NULL)
	    return(NULL);

	pv->textlines = NULL;

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint  border_minor = 2,
			border_major = 5;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3;

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;




	}
	 
	return(CONVPARM(pv));
}

static void ConvParmTextDelete(gpointer data)
{
	convparm_text_struct *pv = CONVPARM_TEXT(data);
	if(!CONVPARM_IS_TEXT(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->prev_btn);
	GTK_WIDGET_DESTROY(pv->entry);
	GTK_WIDGET_DESTROY(pv->next_btn);
	GTK_WIDGET_DESTROY(pv->add_btn);
	GTK_WIDGET_DESTROY(pv->edit_btn);
	GTK_WIDGET_DESTROY(pv->remove_btn);
	g_list_foreach(
	    pv->textlines, (GFunc)ConvParmTextLineDelete, NULL
	);
	g_list_free(pv->textlines);
}

void ConvParmTextSet(convparm_struct *p, GList *textlines)
{
	GList *glist;
	const convparm_textline_struct *l_src;
	convparm_textline_struct *l_tar;
	convparm_text_struct *pv = CONVPARM_TEXT(p);
	if(!CONVPARM_IS_TEXT(pv))
	    return;

	g_list_foreach(                    
	    pv->textlines, (GFunc)ConvParmTextLineDelete, NULL
	);                                                    
	g_list_free(pv->textlines);
	pv->textlines = NULL;
	pv->sel_textline = 0;

	for(glist = textlines; glist != NULL; glist = g_list_next(glist))
	{
	    l_src = CONVPARM_TEXTLINE(glist->data);
	    if(l_src == NULL)
		continue;

	    l_tar = CONVPARM_TEXTLINE(g_malloc0(
		sizeof(convparm_textline_struct)
	    ));
	    if(l_tar != NULL)
	    {
		l_tar->text = STRDUP(l_src->text);
		l_tar->font_name = STRDUP(l_src->font_name);
		l_tar->x = l_src->x;
		l_tar->y = l_src->y;
		l_tar->angle = l_src->angle;
		pv->textlines = g_list_append(
		    pv->textlines, l_tar
		);
	    }
	}


}

GList *ConvParmTextGet(convparm_struct *p)
{
	convparm_text_struct *pv = CONVPARM_TEXT(p);
	if(!CONVPARM_IS_TEXT(pv))                    
	    return(NULL);

	return(pv->textlines);
}

void ConvParmTextLineDelete(convparm_textline_struct *l)
{
	if(l == NULL)
	    return;
	g_free(l->text);
	g_free(l->font_name);
	g_free(l);
}


/* Frame */
convparm_struct *ConvParmFrameNew(GtkWidget *parent)
{
	convparm_frame_struct *pv = CONVPARM_FRAME(ConvParmNew(
	    CONVPARM_TYPE_FRAME,
	    sizeof(convparm_frame_struct),
	    (guint8 **)icon_convparm_frame_20x20_xpm,
	    "Frame",
	    parent,
	    ConvParmFrameDelete
	));
	if(pv == NULL)
	    return(NULL);

	memset(&pv->frameinfo, 0x00, sizeof(convparm_frameinfo_struct));

	parent = CONVPARM_TOPLEVEL(pv);
	if(parent != NULL)
	{
	    const gint  border_minor = 2,
			border_major = 5;
	    GtkAdjustment *adj;
	    GtkWidget *w, *parent2, *parent3;

	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_signal_connect(
		GTK_OBJECT(CONVPARM_TB(pv)), "toggled",
		GTK_SIGNAL_FUNC(ConvParmToggleSensitiveCB), w
	    );
	    gtk_widget_show(w);
	    parent2 = w;
 
	    /* Width */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("W:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 5.0f, 0.0f
	    );
	    pv->width_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Height */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("H:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f, 0.0f, (gfloat)((guint16)-1), 1.0f, 5.0f, 0.0f
	    );
	    pv->height_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Inner Bevel */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("Inner Bevel:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		(gfloat)((guint16)-1) / -2,
		(gfloat)((guint16)-1) / 2,
		1.0f, 5.0f, 0.0f
	    );
	    pv->inner_bevel_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    /* Outer Bevel */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* GtkLabel */
	    w = gtk_label_new("Outer Bevel:");
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* GtkSpinButton */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		(gfloat)((guint16)-1) / -2,
		(gfloat)((guint16)-1) / 2,
		1.0f, 5.0f, 0.0f
	    );
	    pv->outer_bevel_spin = w = gtk_spin_button_new(adj, 1.0f, 0);
	    gtk_widget_set_usize(w, 60, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    GUIEditableEndowPopupMenu(w, 0);
	    gtk_widget_show(w);

	    pv->color_btn = w = CSDColorButtonNewSimple(NULL, -1);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	} 

	return(CONVPARM(pv));
}

static void ConvParmFrameDelete(gpointer data)
{
	convparm_frame_struct *pv = CONVPARM_FRAME(data);
	if(!CONVPARM_IS_FRAME(pv))
	    return;

	GTK_WIDGET_DESTROY(pv->width_spin);
	GTK_WIDGET_DESTROY(pv->height_spin);
	GTK_WIDGET_DESTROY(pv->inner_bevel_spin);
	GTK_WIDGET_DESTROY(pv->outer_bevel_spin);
	GTK_WIDGET_DESTROY(pv->color_btn);
}

void ConvParmFrameSet(
	convparm_struct *p, const convparm_frameinfo_struct *f
)
{
	const GdkColor *c;
	csd_color_struct csd_c;
	convparm_frame_struct *pv = CONVPARM_FRAME(p);
	if(!CONVPARM_IS_FRAME(pv) || (f == NULL))
	    return;

	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->width_spin), (gfloat)f->width
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->height_spin), (gfloat)f->height
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->inner_bevel_spin), (gfloat)f->inner_bevel
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(pv->outer_bevel_spin), (gfloat)f->outer_bevel
	);
	c = &f->c;
	csd_c.a = 1.0f;
	csd_c.r = (gfloat)c->red / (gfloat)((gushort)-1);
	csd_c.g = (gfloat)c->green / (gfloat)((gushort)-1);
	csd_c.b = (gfloat)c->blue / (gfloat)((gushort)-1); 
	CSDColorButtonSetColor(pv->color_btn, &csd_c);
}

const convparm_frameinfo_struct *ConvParmFrameGet(convparm_struct *p)
{
	convparm_frameinfo_struct *f;
	csd_color_struct csd_c;
	convparm_frame_struct *pv = CONVPARM_FRAME(p);
	if(!CONVPARM_IS_FRAME(pv))
	    return(NULL);

	f = &pv->frameinfo;

	f->width = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->width_spin)
	);
	f->height = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->height_spin)
	);
	f->inner_bevel = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->inner_bevel_spin)
	);
	f->outer_bevel = gtk_spin_button_get_value_as_int(
	    GTK_SPIN_BUTTON(pv->outer_bevel_spin)
	);
	CSDColorButtonGetColor(pv->color_btn, &csd_c);
	GDK_COLOR_SET_COEFF(&f->c, csd_c.r, csd_c.g, csd_c.b);

	return(f);
}
