/*
			    Conversion Parameters
 */

#ifndef CONVPARM_H
#define CONVPARM_H

#include <gtk/gtk.h>


typedef struct _convparm_struct		convparm_struct;
typedef struct _convparm_sizeinfo_struct	convparm_sizeinfo_struct;
typedef struct _convparm_textline_struct	convparm_textline_struct;
typedef struct _convparm_frameinfo_struct	convparm_frameinfo_struct;


/*
 *	Conversion Parameter Types:
 */
typedef enum {
	CONVPARM_TYPE_FMT,
	CONVPARM_TYPE_COMPRESSION,
	CONVPARM_TYPE_GAMMA,
	CONVPARM_TYPE_COLORBG,
	CONVPARM_TYPE_MIRROR,
	CONVPARM_TYPE_CROP,
	CONVPARM_TYPE_ROTATE,
	CONVPARM_TYPE_SIZE,
	CONVPARM_TYPE_TEXT,
	CONVPARM_TYPE_FRAME
} convparm_type;


/*
 *	Conversion Parameter Flags:
 */
typedef enum {
	CONVPARM_ENABLED	= (1 << 0),
	CONVPARM_NO_DISABLE	= (1 << 1)
} convparm_flags;


/*
 *	Conversion Parameter:
 */
struct _convparm_struct {
	convparm_type	type;
	convparm_flags	flags;
	gint		freeze_count;
	gchar		*name;

	GtkWidget	*toplevel,
			*tb,
			*icon_pm,
			*label;

	void		(*destroy_func)(
		gpointer		/* Data */
	);
	gpointer	destroy_data;

	void		(*status_mesg_func)(
		const gchar *,		/* Message */
		gpointer		/* Data */
	);
	gpointer	status_mesg_data;

	void		(*help_func)(
		gpointer		/* Data */
	);
	gpointer	help_data;

};
#define CONVPARM(p)		((convparm_struct *)(p))
#define CONVPARM_TOPLEVEL(p)	(((p) != NULL) ? \
 (CONVPARM(p)->toplevel) : NULL)
#define CONVPARM_TB(p)		(((p) != NULL) ? \
 (CONVPARM(p)->tb) : NULL)
#define CONVPARM_IS_ENABLED(p)		(((p) != NULL) ? \
 (CONVPARM(p)->flags & CONVPARM_ENABLED) : FALSE)
#define CONVPARM_IS_NO_DISABLE(p)	(((p) != NULL) ? \
 (CONVPARM(p)->flags & CONVPARM_NO_DISABLE) : FALSE)

#define CONVPARM_FREEZE_COUNT(p)	(((p) != NULL) ? \
 (CONVPARM(p)->freeze_count) : 0)
#define CONVPARM_IS_FROZEN(p)		( \
 CONVPARM_FREEZE_COUNT(p) > 0)
#define CONVPARM_FREEZE(p)	{ \
 if((p) != NULL) CONVPARM(p)->freeze_count++;	\
}
#define CONVPARM_THAW(p)	{ \
 if((p) != NULL) CONVPARM(p)->freeze_count--;	\
}


/*
 *	Conversion Parameter type check macros:
 */
#define CONVPARM_IS_COMPRESSION(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_COMPRESSION) : FALSE)
#define CONVPARM_IS_FMT(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_FMT) : FALSE)
#define CONVPARM_IS_COLORBG(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_COLORBG) : FALSE)
#define CONVPARM_IS_GAMMA(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_GAMMA) : FALSE)
#define CONVPARM_IS_CROP(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_CROP) : FALSE)
#define CONVPARM_IS_MIRROR(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_MIRROR) : FALSE)
#define CONVPARM_IS_ROTATE(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_ROTATE) : FALSE)
#define CONVPARM_IS_SIZE(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_SIZE) : FALSE)
#define CONVPARM_IS_TEXT(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_TEXT) : FALSE)
#define CONVPARM_IS_FRAME(p)	(((p) != NULL) ? \
 (CONVPARM(p)->type == CONVPARM_TYPE_FRAME) : FALSE)


/*
 *	Size Info:
 */
struct _convparm_sizeinfo_struct {
	gint		width,
			height;
	gboolean	maintain_aspect;
	gboolean	smaller;	/* If maintaining aspect, then make
					 * sure that the image is smaller
					 * than the specified width and
					 * height, otherwise the larger of
					 * the width or height will be used */
};
#define CONVPARM_SIZEINFO(p)	((convparm_sizeinfo_struct *)(p))

/*
 *	Text Line:
 */
struct _convparm_textline_struct {
	gchar		*text,		/* Text to draw */
			*font_name;	/* Font name */
	gint		x,
			y;
	gfloat		angle;
};
#define CONVPARM_TEXTLINE(p)	((convparm_textline_struct *)(p))

/*
 *	Frame Info:
 */
struct _convparm_frameinfo_struct {
	gint		width,
			height,
			inner_bevel,
			outer_bevel;
	GdkColor	c;
};
#define CONVPARM_FRAMEINFO(p)	((convparm_frameinfo_struct *)(p))


/* Basics */
extern void ConvParmDelete(convparm_struct *p);
extern void ConvParmShow(convparm_struct *p);
extern void ConvParmHide(convparm_struct *p);
extern void ConvParmEnable(convparm_struct *p);  
extern void ConvParmDisable(convparm_struct *p);
extern convparm_struct *ConvParmListMatchType(
	GList *convparm_list, convparm_type type
);

/* Format */
extern convparm_struct *ConvParmFmtNew(GtkWidget *parent);
extern void ConvParmFmtSet(convparm_struct *p, const gchar *fmt);
extern const gchar *ConvParmFmtGet(convparm_struct *p);
extern gboolean ConvParmFmtCompare(const gchar *s1, const gchar *s2);

/* Compression */
extern convparm_struct *ConvParmCompressionNew(GtkWidget *parent);
extern void ConvParmCompressionSet(convparm_struct *p, gfloat compression);
extern gfloat ConvParmCompressionGet(convparm_struct *p);

/* Background Color */
extern convparm_struct *ConvParmColorBGNew(GtkWidget *parent); 
extern void ConvParmColorBGSet(convparm_struct *p, const GdkColor *c);
extern const GdkColor *ConvParmColorBGGet(convparm_struct *p);          

/* Gamma */
extern convparm_struct *ConvParmGammaNew(GtkWidget *parent);
extern void ConvParmGammaSet(convparm_struct *p, gfloat g);
extern gfloat ConvParmGammaGet(convparm_struct *p);

/* Mirror */
extern convparm_struct *ConvParmMirrorNew(GtkWidget *parent);
extern void ConvParmMirrorSet(convparm_struct *p, GtkOrientation orientation);
extern GtkOrientation ConvParmMirrorGet(convparm_struct *p);

/* Crop */
extern convparm_struct *ConvParmCropNew(GtkWidget *parent);
extern void ConvParmCropSet(convparm_struct *p, const GdkRectangle *rect);
extern const GdkRectangle *ConvParmCropGet(convparm_struct *p);

/* Rotate */
extern convparm_struct *ConvParmRotateNew(GtkWidget *parent);
extern void ConvParmRotateSet(convparm_struct *p, gfloat deg);
extern gfloat ConvParmRotateGet(convparm_struct *p);               

/* Size */
extern convparm_struct *ConvParmSizeNew(GtkWidget *parent);
extern void ConvParmSizeSet(
	convparm_struct *p, const convparm_sizeinfo_struct *s
);
extern const convparm_sizeinfo_struct *ConvParmSizeGet(convparm_struct *p);

/* Text */

extern void ConvParmTextLineDelete(convparm_textline_struct *l);


/* Frame */
extern convparm_struct *ConvParmFrameNew(GtkWidget *parent);
extern void ConvParmFrameSet(
	convparm_struct *p, const convparm_frameinfo_struct *f
);
extern const convparm_frameinfo_struct *ConvParmFrameGet(convparm_struct *p);


#endif	/* CONVPARM_H */
