#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif
#include "../include/disk.h"
#include "guiutils.h"
#include "obj.h"
#include "winpreview.h"
#include "win.h"    
#include "wincb.h"
#include "windnd.h"
#include "winopcb.h"
#include "core.h"


gint WinDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

gint WinOPIDEnterEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);
gint WinOPIDLeaveEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
);

void WinToolBarItemCB(
	ToolBarItem *item, gint id, gpointer data
);
void WinToolBarItemEnterCB(
	ToolBarItem *item, gint id, gpointer data
);
void WinToolBarItemLeaveCB(
	ToolBarItem *item, gint id, gpointer data
);

gint WinEntryEntryCompletePathKeyCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
);

gint WinSrcListEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void WinSrcListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
);
void WinSrcListClickColumnCB(
	GtkCList *clist, gint column, gpointer data             
);
void WinSrcListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);
void WinSrcListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
);

gint WinTarListEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
void WinTarListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
);
void WinTarListClickColumnCB(
	GtkCList *clist, gint column, gpointer data
);
void WinTarListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
); 
void WinTarListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
); 

void WinAnyChangedCB(GtkWidget *widget, gpointer data);

void WinEntryClearCB(GtkWidget *widget, gpointer data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Win toplevel GtkWindow "delete_event" signal callback.
 */
gint WinDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	win_struct *win = WIN(data);
	if(win == NULL)
	    return(FALSE);

	if(win->freeze_count > 0)
	    return(FALSE);

	WinCloseCB(widget, data);

	return(TRUE);
}


/*
 *	Win OPID "enter_notify_event" signal callback.
 */
gint WinOPIDEnterEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	win_struct *win;
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return(FALSE);

	win = opid->win;
	if(win == NULL)
	    return(FALSE);

	if(win->freeze_count > 0)
	    return(FALSE);

	if(!STRISEMPTY(opid->tooltip))
	    WinStatusMessage(
		win,
		opid->tooltip,
		FALSE
	    );

	return(TRUE);
}

/*
 *	Win OPID "leave_notify_event" signal callback.
 */
gint WinOPIDLeaveEventCB(
	GtkWidget *widget, GdkEventCrossing *crossing, gpointer data
)
{
	win_struct *win;
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return(FALSE);

	win = opid->win;
	if(win == NULL)
	    return(FALSE);

	if(win->freeze_count > 0)
	    return(FALSE);

	WinStatusMessage(win, NULL, FALSE);

	return(TRUE);
}


/*
 *	Win Tool Bar Item callback.
 */
void WinToolBarItemCB(
	ToolBarItem *item, gint id, gpointer data
)
{
	win_struct *win;
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return;

	win = opid->win;
	if(win == NULL)
	    return;

	if(win->freeze_count > 0)
	    return;

	if(opid->func_cb != NULL)
	    opid->func_cb(
		ToolBarItemGetWidget(item),	/* GtkWidget */
		win				/* Win */
	    );
}

/*
 *	Win Tool Bar Item enter callback.
 */
void WinToolBarItemEnterCB(
	ToolBarItem *item, gint id, gpointer data
)
{
	win_struct *win;
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return;

	win = opid->win;
	if(win == NULL)
	    return;

	if(win->freeze_count > 0)
	    return;

	WinOPIDEnterEventCB(
	    ToolBarItemGetWidget(item),
	    NULL,
	    opid
	);
}

/*
 *	Win Tool Bar Item leave callback.
 */
void WinToolBarItemLeaveCB(
	ToolBarItem *item, gint id, gpointer data
)
{
	win_struct *win;
	win_opid_struct *opid = WIN_OPID(data);
	if(opid == NULL)
	    return;

	win = opid->win;
	if(win == NULL)
	    return;

	if(win->freeze_count > 0)
	    return;

	WinOPIDLeaveEventCB(
	    ToolBarItemGetWidget(item),
	    NULL,
	    opid
	);
}


/*
 *	GtkEntry complete path "key_press_event" or "key_release_event"
 *	signal callback.
 */
gint WinEntryEntryCompletePathKeyCB(
	GtkWidget *widget, GdkEventKey *key, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	GtkEntry *entry;
	win_struct *win = WIN(data);
	if((widget == NULL) || (key == NULL) || (win == NULL))
	    return(status);

	press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
							    
	/* Given widget must have or be a GtkEntry */                        
	if(GTK_IS_COMBO(widget))
	{
	    GtkCombo *combo = GTK_COMBO(widget);
	    entry = GTK_ENTRY(combo->entry);
	}
	else if(GTK_IS_ENTRY(widget))
	{
	    entry = GTK_ENTRY(widget);
	}
	else
	{
	    return(status);
	}

#define SIGNAL_EMIT_STOP	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}

	switch(key->keyval)
	{
	  case GDK_Tab:
	    if(press)
	    {
		gchar *s = STRDUP(gtk_entry_get_text(entry));
		if(s != NULL)
		{
		    gint status;
							    
		    s = CompletePath(s, &status);
		    gtk_entry_set_text(entry, (s != NULL) ? s : "");
		    gtk_entry_set_position(entry, -1);
		    g_free(s);

		    switch(status)
		    {
		      case COMPLETE_PATH_NONE:
		      case COMPLETE_PATH_AMBIGUOUS:
			gdk_beep();
			break;
		    }
		}
	    }
	    SIGNAL_EMIT_STOP
	    status = TRUE;
	    break;
	}

#undef SIGNAL_EMIT_STOP

	return(status);
}


/*
 *	Win list event signal callback.
 */
gint WinSrcListEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint status = FALSE;
	gboolean press;
	guint keyval, state;
	GdkEventFocus *focus;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
	    return(status);

	if(win->freeze_count > 0)
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->in)
		win->focus_clist = widget;
	    WinUpdate(win);
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    press = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	    keyval = key->keyval;
	    state = key->state;
#define DO_STOP_KEY_SIGNAL_EMIT	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}
	    switch(keyval)
	    {
	      case GDK_Return:
	      case GDK_KP_Enter:
	      case GDK_ISO_Enter:
		break;

	      case GDK_F3:
		if(press)
		    WinAddCB(win->add_mi, win);
		status = TRUE;
		break;

	      case GDK_Insert:
		if(state & GDK_CONTROL_MASK)
		{
		    if(press)
			WinCopyCB(win->copy_mi, win);
		    status = TRUE;
		}
		else if(state & GDK_SHIFT_MASK)
		{
		    if(press)
			WinPasteCB(win->copy_mi, win);
		    status = TRUE;
		}
		break;

	      case GDK_Delete:
		if(press)
		    WinRemoveCB(win->remove_mi, win);
		status = TRUE;
		break;
	    }
#undef DO_STOP_KEY_SIGNAL_EMIT
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		break;

	      case GDK_BUTTON2:
		break;

	      case GDK_BUTTON3:
		if(win->src_list_menu != NULL)
		{
		    gtk_menu_popup(
			GTK_MENU(win->src_list_menu),
			NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		    status = TRUE;
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;

	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;

	    break;
	}

	return(status);
}

/*
 *	Win GtkCList "resize_column" signal callback.
 */
void WinSrcListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
)
{
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	WinUpdate(win);
}

/*
 *	Win GtkCList "click_column" signal callback.
 */
void WinSrcListClickColumnCB(
	GtkCList *clist, gint column, gpointer data             
)
{
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	WinUpdate(win);
}

/*
 *	Win GtkCList "select_row" signal callback.
 */
void WinSrcListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	gint nselected;
	GList *glist;
	core_struct *core;
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	glist = clist->selection;
	core = CORE(win->core);
	if(core == NULL)
	    return;

	/* Set the DND drag icon */
	GUIDNDSetDragIconFromCListSelection(clist);

	nselected = (gint)g_list_length(glist);
	if(nselected == 1)
	{
	    obj_struct *obj = OBJ(gtk_clist_get_row_data(
		clist, (gint)glist->data
	    ));
	    if((obj != NULL) ? !STRISEMPTY(obj->full_path) : FALSE)
	    {
		struct stat lstat_buf;
		const gchar *path = obj->full_path;
		if(lstat(path, &lstat_buf))
		    memset(&lstat_buf, 0x00, sizeof(struct stat));

		win->freeze_count++;

#ifdef S_ISREG
		if(S_ISREG(lstat_buf.st_mode))
#else
		if(TRUE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Displaying \"%s\"...",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);

		    WinSetBusy(win, TRUE);
#ifdef HAVE_LIBENDEAVOUR2
		    /* Preview only if specified and not converting */
/* NOTE: Previewing while converting causes a pthread error with
 * ImageMagick, because sometimes this function is called during an
 * ImageMagick progress callback which causes a threading error in
 * ImageMagick
 */
		    if(win->preview_src && !win->processing)
		        WinPreviewSetImageFile(
			    win->src_preview, path,
			    core->edv_ctx
		        );
		    s = g_strdup_printf(
"File \"%s\" selected (%s %s)",
			obj->name,
			edv_str_size_delim(lstat_buf.st_size),
			(lstat_buf.st_size == 1l) ? "byte" : "bytes"
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
#else
		    /* Preview only if specified and not converting */
		    if(win->preview_src && !win->processing)
			WinPreviewSetImageFile(
			    win->src_preview, path
		        );
		    s = g_strdup_printf(
"File \"%s\" selected (%ld %s)",
			obj->name,
			(gulong)lstat_buf.st_size,
			(lstat_buf.st_size == 1l) ? "byte" : "bytes"
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
#endif
		    WinSetBusy(win, FALSE);
		}
#ifdef S_ISDIR
		else if(S_ISDIR(lstat_buf.st_mode))
#else
		else if(FALSE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Directory \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}
#ifdef S_ISLNK
		else if(S_ISLNK(lstat_buf.st_mode))
#else
		else if(FALSE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Link \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}
		else
		{
		    gchar *s = g_strdup_printf(
			"Object \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}

		win->freeze_count--;
	    }
	}
	else if(nselected == 0)
	{
	    win->freeze_count++;
	    WinStatusMessage(win, "No objects selected", TRUE);
	    win->freeze_count--;
	}
	else
	{
#ifdef HAVE_LIBENDEAVOUR2
	    gchar *s = g_strdup_printf(
		"%s objects selected",
		edv_str_size_delim((gulong)nselected)
	    );
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
#else
	    gchar *s = g_strdup_printf(
		"%i objects selected",
		nselected
	    );
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
#endif
	}

	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column */
		0.5f, 0.0f	/* Row, column */
	    );

	WinUpdate(win);

	/* Double clicked? */
	if(event != NULL)
	{
	    if(event->type == GDK_2BUTTON_PRESS)
	    {
		GdkEventButton *button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON1:
		    if(button->state & GDK_SHIFT_MASK)
		    {
			/* Open */
			WinOpenCB(win->open_mi, win);
		    }
		    else
		    {
			/* Convert */
			WinConvertSelectedCB(win->convert_selected_mi, win);
		    }
		    break;
		}
	    }
	}
}

/*
 *	Win GtkCList "unselect_row" signal callback.
 */
void WinSrcListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	gint nselected;
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	nselected = (gint)g_list_length(clist->selection);
	if(nselected > 0)
	{
#ifdef HAVE_LIBENDEAVOUR2
	    gchar *s = g_strdup_printf(
		"%s object%s selected",
		edv_str_size_delim((gulong)nselected),
		(nselected == 1) ? "" : "s"
	    );
#else
	    gchar *s = g_strdup_printf(
		"%i object%s selected",
		nselected,
		(nselected == 1) ? "" : "s"
	    );
#endif
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
	}
	else if(nselected == 0)
	{
	    WinPreviewClear(win->src_preview);
	    win->freeze_count++;
	    WinStatusMessage(win, "No objects selected", TRUE);
	    win->freeze_count--;
	}

	WinUpdate(win);
}


/*
 *      Win list event signal callback.
 */
gint WinTarListEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gboolean press;
	guint keyval, state;
	GdkEventFocus *focus;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	win_struct *win = WIN(data);
	if((widget == NULL) || (event == NULL) || (win == NULL))
	    return(status);

	if(win->freeze_count > 0)
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_FOCUS_CHANGE:
	    focus = (GdkEventFocus *)event;
	    if(focus->in)
		win->focus_clist = widget;
	    WinUpdate(win);
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    key = (GdkEventKey *)event;
	    press = (event->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	    keyval = key->keyval;
	    state = key->state;
#define DO_STOP_KEY_SIGNAL_EMIT	{		\
 gtk_signal_emit_stop_by_name(			\
  GTK_OBJECT(widget),				\
  press ?					\
   "key_press_event" : "key_release_event"	\
 );						\
}
	    switch(keyval)
	    {
	      case GDK_Return:
	      case GDK_KP_Enter:
	      case GDK_ISO_Enter:
		break;
	    }
#undef DO_STOP_KEY_SIGNAL_EMIT
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    switch(button->button)
	    {
	      case GDK_BUTTON1:
		break;
		     
	      case GDK_BUTTON2:
		break;

	      case GDK_BUTTON3:
		if(win->tar_list_menu != NULL)
		{
		    gtk_menu_popup(
			GTK_MENU(win->tar_list_menu),
			NULL, NULL,
			NULL, NULL,
			button->button, button->time
		    );
		    status = TRUE;
		}
		break;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;

	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;

	    break;
	}
	 
	return(status);
}

/*
 *	Win GtkCList "resize_column" signal callback.
 */
void WinTarListResizeColumnCB(
	GtkCList *clist, gint column, gint width, gpointer data
)
{
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	WinUpdate(win);
}

/*
 *	Win GtkCList "click_column" signal callback.
 */
void WinTarListClickColumnCB(
	GtkCList *clist, gint column, gpointer data
)
{
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	WinUpdate(win);
}

/*
 *      Win GtkCList "select_row" signal callback.
 */
void WinTarListSelectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	gint nselected;
	GList *glist;
	core_struct *core;
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	glist = clist->selection;
	core = CORE(win->core);
	if(core == NULL)
	    return;

	nselected = (gint)g_list_length(glist);

	/* Set the DND drag icon */
	GUIDNDSetDragIconFromCListSelection(clist);

	if(nselected == 1)
	{
	    obj_struct *obj = OBJ(gtk_clist_get_row_data(
		clist, (gint)glist->data
	    ));
	    if((obj != NULL) ? !STRISEMPTY(obj->full_path) : FALSE)
	    {
		struct stat lstat_buf;
		const gchar *path = obj->full_path;
		if(lstat(path, &lstat_buf))
		    memset(&lstat_buf, 0x00, sizeof(struct stat));

		win->freeze_count++;

#ifdef S_ISREG
		if(S_ISREG(lstat_buf.st_mode))
#else
		if(TRUE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Displaying \"%s\"...",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);

		    WinSetBusy(win, TRUE);
#ifdef HAVE_LIBENDEAVOUR2
		    /* Preview only if specified and not converting */
		    if(win->preview_tar && !win->processing)
			WinPreviewSetImageFile(
			    win->tar_preview, path,
			    core->edv_ctx
		        );
		    s = g_strdup_printf(
"File \"%s\" selected (%s %s)",
			obj->name,
			edv_str_size_delim(lstat_buf.st_size),
			(lstat_buf.st_size == 1l) ? "byte" : "bytes"
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
#else
		    /* Preview only if specified and not converting */
		    if(win->preview_tar && !win->processing)
			WinPreviewSetImageFile(
			    win->tar_preview, path
		        );
		    s = g_strdup_printf(
"File \"%s\" selected (%ld %s)",
			obj->name,
			(gulong)lstat_buf.st_size,
			(lstat_buf.st_size == 1l) ? "byte" : "bytes"
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
#endif

		    WinSetBusy(win, FALSE);
		}
#ifdef S_ISDIR
		else if(S_ISDIR(lstat_buf.st_mode))
#else
		else if(FALSE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Directory \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}
#ifdef S_ISLNK
		else if(S_ISLNK(lstat_buf.st_mode))
#else
		else if(FALSE)
#endif
		{
		    gchar *s = g_strdup_printf(
			"Link \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}
		else
		{
		    gchar *s = g_strdup_printf(
			"Object \"%s\" selected",
			obj->name
		    );
		    WinStatusMessage(win, s, TRUE);
		    g_free(s);
		}

		win->freeze_count--;
	    }
	}
	else if(nselected == 0)
	{
	    win->freeze_count++;
	    WinStatusMessage(win, "No objects selected", TRUE);
	    win->freeze_count--;
	}
	else
	{
#ifdef HAVE_LIBENDEAVOUR2
	    gchar *s = g_strdup_printf(
		"%s objects selected",
		edv_str_size_delim((gulong)nselected)
	    );
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
#else
	    gchar *s = g_strdup_printf(
		"%i objects selected",
		nselected
	    );
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
#endif
	}

	if(gtk_clist_row_is_visible(clist, row) !=
	    GTK_VISIBILITY_FULL
	)
	    gtk_clist_moveto(
		clist,
		row, -1,	/* Row, column */
		0.5f, 0.0f	/* Row, column */
	    );

	WinUpdate(win);

	/* Double clicked? */
	if(event != NULL)
	{
	    if(event->type == GDK_2BUTTON_PRESS)
	    {
		GdkEventButton *button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON1:
		    /* Open */
		    WinOpenCB(win->open_mi, win);
		    break;
		}
	    }
	}
}

/*
 *	Win GtkCList "unselect_row" signal callback.
 */
void WinTarListUnselectRowCB(
	GtkCList *clist, gint row, gint column, GdkEvent *event,
	gpointer data
)
{
	gint nselected;
	win_struct *win = WIN(data);
	if((clist == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	nselected = (gint)g_list_length(clist->selection);
	if(nselected > 0)
	{
#ifdef HAVE_LIBENDEAVOUR2
	    gchar *s = g_strdup_printf(
		"%s object%s selected",
		edv_str_size_delim((gulong)nselected),
		(nselected == 1) ? "" : "s"
	    );
#else
	    gchar *s = g_strdup_printf(
		"%i object%s selected",
		nselected,
		(nselected == 1) ? "" : "s"
	    );
#endif
	    win->freeze_count++;
	    WinStatusMessage(win, s, TRUE);
	    g_free(s);
	    win->freeze_count--;
	}
	else if(nselected == 0)
	{
	    WinPreviewClear(win->tar_preview);
	    win->freeze_count++;
	    WinStatusMessage(win, "No objects selected", TRUE);
	    win->freeze_count--;
	}

	WinUpdate(win);
}


/*
 *	Win any GtkWidget "changed" or "toggled" signal callback.
 */
void WinAnyChangedCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if((widget == NULL) || (win == NULL))
	    return;

	if(win->freeze_count > 0)
	    return;

	if(widget == win->tar_dir_entry)
	{
	    gchar *s = STRDUP(gtk_entry_get_text(GTK_ENTRY(widget)));
	    WinSetTarDir(win, s);
	    g_free(s);
	}
}


/*
 *	Clear GtkEntry callback.
 *
 *	The data is assumed to be a GtkEntry.
 */
void WinEntryClearCB(GtkWidget *widget, gpointer data)
{
	GtkEntry *entry = (GtkEntry *)data;
	if(entry != NULL)
	    gtk_entry_set_text(entry, "");
}
