#include <gtk/gtk.h>

#include "guiutils.h"

#include "obj.h"
#include "winpreview.h"
#include "win.h"
#include "winlist.h"
#include "windnd.h"
#include "config.h"


void WinCListDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);
void WinCListDragDataGetCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);
void WinCListDragDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
);

void WinPreviewDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	GtkCList "drag_data_received" signal callback.
 */
void WinCListDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	gint row, column;
	GtkCList *clist;
	win_struct *win = WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
	    return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
	    return;

	/* Calculate the row and column based on the drop coordinates */
	if(!gtk_clist_get_selection_info(
	    clist,
	    x,    
	    y - ((clist->flags & GTK_CLIST_SHOW_TITLES) ?
		clist->column_title_area.height +
		clist->column_title_area.y : 0),
	    &row, &column
	))
	{
	    row = -1;
	    column = 0;
	}

	WinSetBusy(win, TRUE);

	/* Handle by data type */
	if((info == DND_INFO_TEXT_PLAIN) ||
	   (info == DND_INFO_TEXT_URI_LIST) ||
	   (info == DND_INFO_STRING)
	)
	{
	    gint new_row, insert_row = row;
	    GList *glist;
	    obj_struct *obj;

	    /* Parse the URL string list into a list of objects */
	    GList *obj_list = ObjDDEBufferParse(
		(const guint8 *)selection_data->data,
		selection_data->length
	    );

	    /* Add each object to the list */
	    for(glist = obj_list; glist != NULL; glist = g_list_next(glist))
	    {
		obj = OBJ(glist->data);
		if(obj == NULL)
		    continue;

		if(widget == win->src_clist)
		    new_row = WinSrcListInsert(
			win, insert_row, ObjCopy(obj)
		    );
		else if(widget == win->tar_clist)
		    new_row = WinTarListInsert(
			win, insert_row, ObjCopy(obj)
		    );
		else
		    new_row = -1;
		if(new_row > -1)
		{
		    if(insert_row > -1)
			insert_row++;
		}
	    }

	    /* Delete the objects list */
	    g_list_foreach(obj_list, (GFunc)ObjDelete, NULL);
	    g_list_free(obj_list);
	}

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	GtkCList "drag_data_get" signal callback.
 */
void WinCListDragDataGetCB(
	GtkWidget *widget, GdkDragContext *dc,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
)
{
	gboolean data_sent = FALSE;
	gint row, buf_len = 0;
	guint8 *buf = NULL;
	GList *glist;
	GtkCList *clist;
	win_struct *win = WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
	    return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
	    return;

	/* Iterate through selected rows */
	for(glist = clist->selection;
	    glist != NULL;
	    glist = g_list_next(glist)
	)
	{
	    row = (gint)glist->data;
	    buf = ObjDDEBufferAppend(
		buf, &buf_len,
		OBJ(gtk_clist_get_row_data(clist, row))
	    );
	}
	if(buf != NULL)
	{
	    gtk_selection_data_set( 
		selection_data,
		GDK_SELECTION_TYPE_STRING,
		8,                      /* Bits Per Character */
		buf, buf_len
	    );
	    data_sent = TRUE;
	    g_free(buf);
	}

	/* If no data was sent out then send an error response */
	if(!data_sent)
	{
	    const gchar *s = "Error";
	    gtk_selection_data_set(
		selection_data,
		GDK_SELECTION_TYPE_STRING,
		8,			/* Bits Per Character */
		s,			/* Data */
		STRLEN(s)		/* Length */
	    ); 
	    data_sent = TRUE;
	}
}

/*
 *	GtkCList "drag_data_delete" signal callback.
 */
void WinCListDragDataDeleteCB(
	GtkWidget *widget, GdkDragContext *dc, gpointer data
)
{
	GtkCList *clist;
	win_struct *win = WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
	    return;

	clist = GTK_IS_CLIST(widget) ? GTK_CLIST(widget) : NULL;
	if(clist == NULL)
	    return;

	WinSetBusy(win, TRUE);

	gtk_clist_freeze(clist);
	while(clist->selection_end != NULL)
	    gtk_clist_remove(
		clist, (gint)clist->selection_end->data
	    );
	gtk_clist_thaw(clist);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Preview "drag_data_received" signal callback.
 */
void WinPreviewDragDataReceivedCB(
	GtkWidget *widget, GdkDragContext *dc,
	gint x, gint y,
	GtkSelectionData *selection_data, guint info, guint t,
	gpointer data
					 
)
{
	win_struct *win = WIN(data);
	if((widget == NULL) || (dc == NULL) || (win == NULL))
	    return;

	WinSetBusy(win, TRUE);

	/* Handle by data type */
	if((info == DND_INFO_TEXT_PLAIN) ||
	   (info == DND_INFO_TEXT_URI_LIST) ||
	   (info == DND_INFO_STRING)
	)
	{
	    GList *glist;
	    obj_struct *obj;

	    /* Parse the URL string list into a list of objects */
	    GList *obj_list = ObjDDEBufferParse(
		(const guint8 *)selection_data->data,
		selection_data->length
	    );

	    /* Add each object to the list */
	    for(glist = obj_list; glist != NULL; glist = g_list_next(glist))
	    {
		obj = OBJ(glist->data);
		if(obj == NULL)
		    continue;

		if(widget == WinPreviewGetDrawingArea(win->src_preview))
		    WinSrcListAppend(win, ObjCopy(obj));
		else if(widget == WinPreviewGetDrawingArea(win->tar_preview))
		    WinTarListAppend(win, ObjCopy(obj));
	    }

	    /* Delete the objects list */
	    g_list_foreach(obj_list, (GFunc)ObjDelete, NULL);
	    g_list_free(obj_list);
	}

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}
