#ifdef HAVE_IMAGE_MAGICK
# include <stdio.h>
# include <sys/types.h>
# include <wand/magick_wand.h>
#endif
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#ifndef HAVE_LIBENDEAVOUR2
#include "../include/disk.h"
#include "../include/prochandle.h"
#endif

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"

#include "obj.h"
#include "help.h"
#include "winpreview.h"
#include "win.h"
#include "winconv.h"
#include "wincb.h"
#include "winlist.h"
#include "winopcb.h"
#include "core.h"
#include "cfglist.h"
#include "config.h"


#include "images/icon_imgworks_32x32.xpm"


static void WinCListSelectAll(GtkCList *clist);
static void WinCListUnselectAll(GtkCList *clist);
static void WinCListInvertSelection(GtkCList *clist);

void WinCloseCB(GtkWidget *widget, gpointer data);
void WinExitCB(GtkWidget *widget, gpointer data);

void WinNewCB(GtkWidget *widget, gpointer data);
void WinAddCB(GtkWidget *widget, gpointer data);
void WinRemoveCB(GtkWidget *widget, gpointer data);
void WinClearSrcCB(GtkWidget *widget, gpointer data);
void WinClearTarCB(GtkWidget *widget, gpointer data);
void WinShiftUpCB(GtkWidget *widget, gpointer data);
void WinShiftDownCB(GtkWidget *widget, gpointer data);

void WinCutCB(GtkWidget *widget, gpointer data);
void WinCopyCB(GtkWidget *widget, gpointer data);
void WinPasteCB(GtkWidget *widget, gpointer data);

void WinSelectAllCB(GtkWidget *widget, gpointer data);
void WinUnselectAllCB(GtkWidget *widget, gpointer data);
void WinInvertSelectionCB(GtkWidget *widget, gpointer data);

void WinPropertiesCB(GtkWidget *widget, gpointer data);


void WinSetTarDirCB(GtkWidget *widget, gpointer data);

void WinConvertCB(GtkWidget *widget, gpointer data);
void WinConvertSelectedCB(GtkWidget *widget, gpointer data);
void WinStopCB(GtkWidget *widget, gpointer data);
void WinRemoveConvertedCB(GtkWidget *widget, gpointer data);
void WinRecursiveCB(GtkWidget *widget, gpointer data);

void WinOpenCB(GtkWidget *widget, gpointer data);
void WinRefreshCB(GtkWidget *widget, gpointer data);
void WinPreviewSrcCB(GtkWidget *widget, gpointer data);
void WinPreviewSrcShowDetailsCB(GtkWidget *widget, gpointer data);
void WinPreviewTarCB(GtkWidget *widget, gpointer data);
void WinPreviewTarShowDetailsCB(GtkWidget *widget, gpointer data);
void WinViewToolBarCB(GtkWidget *widget, gpointer data);
void WinViewStatusBarCB(GtkWidget *widget, gpointer data);

void WinNewImgWorksCB(GtkWidget *widget, gpointer data);
void WinNewImbrCB(GtkWidget *widget, gpointer data);

void WinHelpContentsCB(GtkWidget *widget, gpointer data);
void WinHelpAboutCB(GtkWidget *widget, gpointer data);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Selects all rows on the specified GtkCList.
 */
static void WinCListSelectAll(GtkCList *clist)
{
	if(clist == NULL)
		return;

	gtk_clist_freeze(clist);
	gtk_clist_select_all(clist);
	gtk_clist_thaw(clist);
}
 
/*
 *	Unselects all rows on the specified GtkCList.
 */
static void WinCListUnselectAll(GtkCList *clist)
{
	if(clist == NULL)
		return;

	gtk_clist_freeze(clist);
	gtk_clist_unselect_all(clist);
	gtk_clist_thaw(clist);
}

/*
 *	Inverts selection on the specified GtkCList.
 */
static void WinCListInvertSelection(GtkCList *clist)
{
	gint i;
	GList *glist;

	if(clist == NULL)
		return;

	glist = (clist->selection != NULL) ?
		g_list_copy(clist->selection) : NULL;

	gtk_clist_freeze(clist);
	for(i = 0; i < clist->rows; i++)
	{
		if(g_list_find(glist, (gpointer)i) != NULL)
			gtk_clist_unselect_row(clist, i, 0);
		else
			gtk_clist_select_row(clist, i, 0);
	}
	gtk_clist_thaw(clist);
						  
	g_list_free(glist);
}


/*
 *	Close.
 */
void WinCloseCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if((win->freeze_count > 0) || win->processing)
		return;

	WinSetBusy(win, TRUE);

	WinSavePositions(win); 
	WinUnmap(win);

	WinSetBusy(win, FALSE);
}

/*
 *	Exit.
 */
void WinExitCB(GtkWidget *widget, gpointer data)
{
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if((win->freeze_count > 0) || win->processing)
		return;

	core = win->core;
	if(core == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Mark the core's close_all_windows to close all windows during
	 * the core's main management check
	 */
	core->close_all_windows = TRUE;

	WinSavePositions(win);
	WinUpdate(win);

	WinSetBusy(win, FALSE);
}


/*
 *	New.
 */
void WinNewCB(GtkWidget *widget, gpointer data)
{
	WinClearSrcCB(widget, data);
	WinAddCB(widget, data);
}

/*
 *	Add.
 */
void WinAddCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint		npaths = 0,
					nftypes = 0;
	gchar **paths_list = NULL;
	GtkWidget *toplevel;
	GtkCList *clist;
	fb_type_struct	**ftypes_list = NULL,
					*ftype = NULL;
	win_struct *win = WIN(data);
	if((win == NULL) || FileBrowserIsQuery())
		return;

	if(win->freeze_count > 0)
		return;

	toplevel = win->toplevel;
	clist = GTK_CLIST(win->src_clist);
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Create the file types list */
	FileBrowserTypeListNew(
		&ftypes_list, &nftypes,
		"*.*",
		"All images"
	);

	/* Query the user to add images */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
		"Add",
		"Add", "Cancel",
		NULL,
		ftypes_list, nftypes,
		&paths_list, &npaths,
		&ftype
	);
	FileBrowserSetTransientFor(NULL);

	/* Got user response? */
	if(response && (npaths > 0))
	{
		gint	i,
					new_row,
					nadded = 0,
					last_new_row = -1;
		gchar *path;
		const gchar *added_path = NULL;
		GList *glist = clist->selection_end;
		gint ins_row = (glist != NULL) ? (gint)glist->data : -1;

		gtk_clist_freeze(clist);

		/* Add each path */
		for(i = 0; i < npaths; i++)
		{
			path = STRDUP(paths_list[i]);
			if(path == NULL)
				continue;

			new_row = WinSrcListInsert(
				win,
				ins_row,
				ObjNewFromPath(path)
			);
			if(new_row > -1)
			{
				last_new_row = new_row;

				if(ins_row > -1)
					ins_row++;

				nadded++;
				added_path = path;
			}

			g_free(path);
		}

		gtk_clist_thaw(clist);

		/* If the row is not visible then scroll to make it visible */
		if(last_new_row > -1)
		{
			if(gtk_clist_row_is_visible(clist, last_new_row) !=
				GTK_VISIBILITY_FULL
			)
				gtk_clist_moveto(
					clist,
					last_new_row, -1,       /* Row, column */
					0.5f, 0.0f              /* Row, column */
				);
		}

		/* Update the status bar */
		if(nadded > 1)
		{
#ifdef HAVE_LIBENDEAVOUR2
			gchar *s = g_strdup_printf(
				"%s objects added",
				edv_str_size_delim((gulong)nadded)
			);
#else
			gchar *s = g_strdup_printf(
				"%i objects added",
				nadded
			);
#endif
			WinStatusMessage(win, s, FALSE);
			g_free(s);
		}
		else if(nadded == 1)
		{
			gchar *s = g_strdup_printf(
				"Object \"%s\" added",
				(added_path != NULL) ? g_basename(added_path) : "untitled"
			);
			WinStatusMessage(win, s, FALSE);
			g_free(s);
		}
		else
		{
			WinStatusMessage(win, "No objects added", FALSE);
		}
	}

	/* Delete file types list */
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Remove.
 */
void WinRemoveCB(GtkWidget *widget, gpointer data)
{
	gint row, nremoved = 0;
	gchar *removed_name = NULL;
	GtkCList *clist;
	obj_struct *obj;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Remove selected paths */
	gtk_clist_freeze(clist);
	while(clist->selection_end != NULL)
	{
		row = (gint)clist->selection_end->data;
		obj = OBJ(gtk_clist_get_row_data(clist, row));
		if(obj != NULL)
		{
			g_free(removed_name);
			removed_name = STRDUP(obj->name);
		}
		nremoved++;
		gtk_clist_remove(clist, row);
	}
	gtk_clist_thaw(clist);

	/* Report status */
	if(nremoved > 1)
	{
#ifdef HAVE_LIBENDEAVOUR2
		gchar *s = g_strdup_printf(
			"%s objects removed",
			edv_str_size_delim((gulong)nremoved)
		);
#else
		gchar *s = g_strdup_printf(
			"%i objects removed",
			nremoved
		);
#endif
		WinStatusMessage(win, s, FALSE);
		g_free(s);
	}
	else if(nremoved == 1)
	{
		gchar *s = g_strdup_printf(
			"Object \"%s\" removed",
			(removed_name != NULL) ? removed_name : "untitled"
		);
		WinStatusMessage(win, s, FALSE);
		g_free(s);
	}
	else
	{
		WinStatusMessage(win, "No objects removed", FALSE);
	}
	g_free(removed_name);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Clear Source.
 */
void WinClearSrcCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	WinSetBusy(win, TRUE);
	WinSrcListClear(win);
	WinPreviewClear(win->src_preview);
	WinStatusMessage(win, "Source list cleared", FALSE);
	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Clear Converted.
 */
void WinClearTarCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	WinSetBusy(win, TRUE);
	WinTarListClear(win);
	WinPreviewClear(win->tar_preview);
	WinStatusMessage(win, "Converted list cleared", FALSE);
	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Shift Up.
 */
void WinShiftUpCB(GtkWidget *widget, gpointer data)
{
	gint row, nrows;
	GList *selection = NULL, *glist;
	GtkCList *clist;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Get copy of selection list, ordered from first to last */
	nrows = clist->rows;
	for(row = 0; row < nrows; row++)
	{
		if(g_list_find(clist->selection, (gpointer)row) != NULL)
			selection = g_list_append(
				selection,
				(gpointer)row
			);
	}

	/* Shift rows up */
	gtk_clist_freeze(clist);
	for(glist = selection; glist != NULL; glist = g_list_next(glist))
	{
		row = (gint)glist->data;
		if(row <= 0)
			break;

		gtk_clist_swap_rows(clist, row, row - 1);
	}
	gtk_clist_thaw(clist);

	/* Delete copy of selection list */
	g_list_free(selection);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Shift Down.
 */
void WinShiftDownCB(GtkWidget *widget, gpointer data)
{
	gint row, nrows;
	GList *selection = NULL, *glist;
	GtkCList *clist;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Get copy of selection list, ordered from last to first */
	nrows = clist->rows;
	for(row = nrows - 1; row >= 0; row--)
	{
		if(g_list_find(clist->selection, (gpointer)row) != NULL)
			selection = g_list_append(
				selection,
				(gpointer)row
			);
	}

	/* Shift rows down */
	gtk_clist_freeze(clist);
	for(glist = selection; glist != NULL; glist = g_list_next(glist))
	{
		row = (gint)glist->data;
		if(row >= (nrows - 1))
			break;

		gtk_clist_swap_rows(clist, row, row + 1);
	}
	gtk_clist_thaw(clist);

	/* Delete copy of selection list */
	g_list_free(selection);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Cut.
 */
void WinCutCB(GtkWidget *widget, gpointer data)
{
	guint8 *buf = NULL;
	gint row, buf_len = 0;
	GList *obj_list, *glist;
	GtkCList *clist;
	obj_struct *obj;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Create a list of selected objects */
	obj_list = NULL;
	for(glist = clist->selection;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		row = (gint)glist->data;
		obj = OBJ(gtk_clist_get_row_data(clist, row));
		if(obj == NULL)
			continue;

		obj_list = g_list_append(obj_list, obj);
		buf = ObjDDEBufferAppend(buf, &buf_len, obj);
	}

	/* Transfer the buffer to the clipboard */
	if(buf != NULL)
		GUIDDESetDirect(
			GTK_WIDGET(clist),
			GDK_SELECTION_PRIMARY,
			GDK_CURRENT_TIME,
			gdk_atom_intern(GUI_TARGET_NAME_STRING, FALSE),
			buf, buf_len
		);

	/* Delete the objects */
	gtk_clist_freeze(clist);
	for(glist = obj_list;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		row = gtk_clist_find_row_from_data(clist, glist->data);
		if(row > -1)
			gtk_clist_remove(clist, row);
	}
	gtk_clist_thaw(clist);

	/* Delete the selected objects list */
	g_list_free(obj_list);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Copy.
 */
void WinCopyCB(GtkWidget *widget, gpointer data)
{
	guint8 *buf = NULL;
	gint row, buf_len = 0;
	GList *glist;
	GtkCList *clist;
	obj_struct *obj;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Create the buffer describing the selected objects */
	for(glist = clist->selection;
		glist != NULL;
		glist = g_list_next(glist)
	)
	{
		row = (gint)glist->data;
		obj = OBJ(gtk_clist_get_row_data(clist, row));
		if(obj == NULL)
			continue;

		buf = ObjDDEBufferAppend(buf, &buf_len, obj);
	}

	/* Transfer the buffer to the clipboard */
	if(buf != NULL)
		GUIDDESetDirect(
			GTK_WIDGET(clist),
			GDK_SELECTION_PRIMARY,
			GDK_CURRENT_TIME,
			gdk_atom_intern(GUI_TARGET_NAME_STRING, FALSE),
			buf, buf_len
		);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Paste.
 */
void WinPasteCB(GtkWidget *widget, gpointer data)
{
	guint8 *buf;
	gint buf_len;
	GtkCList *clist;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	clist = (GtkCList *)win->focus_clist;
	if(clist == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Get buffer from clipboard */
	buf = GUIDDEGetBinary(
		GTK_WIDGET(clist),
		GDK_SELECTION_PRIMARY,
		GDK_CURRENT_TIME,
		&buf_len
	);
	if(buf != NULL) 
	{
		gint row, last_new_row = -1;
		GList *glist, *obj_list;
		obj_struct *obj;

		/* Parse the buffer into a list of objects */
		obj_list = ObjDDEBufferParse(buf, buf_len);

		/* Delete the buffer */
		g_free(buf);

		gtk_clist_freeze(clist);

		/* Get the last selected row */
		glist = clist->selection_end;
		row = (glist != NULL) ? (gint)glist->data : -1;

		/* Insert or append? */
		if(row > -1)
		{
			gint ins_row = row;

			for(glist = obj_list; glist != NULL; glist = g_list_next(glist))
			{
				obj = OBJ(glist->data);
				if(GTK_WIDGET(clist) == win->tar_clist)
					ins_row = WinTarListInsert(
						win, ins_row, ObjCopy(obj)
					);
				else
					ins_row = WinSrcListInsert(
						win, ins_row, ObjCopy(obj)
					);
				if(ins_row < 0)
					break;

				last_new_row = ins_row;

				ins_row++;
			}
		}
		else
		{
			gint new_row;

			for(glist = obj_list; glist != NULL; glist = g_list_next(glist))
			{
				obj = OBJ(glist->data);
				if(GTK_WIDGET(clist) == win->tar_clist)
					new_row = WinTarListAppend(win, ObjCopy(obj));
				else
					new_row = WinSrcListAppend(win, ObjCopy(obj));
				if(new_row < 0)
					break;

				last_new_row = new_row;
			}
		}

		gtk_clist_thaw(clist);

		/* Delete the object list */
		if(obj_list != NULL)
		{
			g_list_foreach(obj_list, (GFunc)ObjDelete, NULL);
			g_list_free(obj_list);
		}

		/* If the row is not visible then scroll to make it visible */
		if(last_new_row > -1)
		{
			if(gtk_clist_row_is_visible(clist, last_new_row) !=
				GTK_VISIBILITY_FULL
			)
				gtk_clist_moveto(
					clist,
					last_new_row, -1,       /* Row, column */
					0.5f, 0.0f              /* Row, column */
				);
		}
	}

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Select All.
 */
void WinSelectAllCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);
	WinCListSelectAll((GtkCList *)win->focus_clist);
	WinSetBusy(win, FALSE);
}

/*
 *	Unselect All.
 */
void WinUnselectAllCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);
	WinCListUnselectAll((GtkCList *)win->focus_clist);
	WinSetBusy(win, FALSE);
}

/*
 *	Invert Selection.
 */
void WinInvertSelectionCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);
	WinCListInvertSelection((GtkCList *)win->focus_clist);
	WinSetBusy(win, FALSE);
}


/*
 *	Properties.
 */
void WinPropertiesCB(GtkWidget *widget, gpointer data)
{
#if defined(HAVE_LIBENDEAVOUR2)
	gint row;
	gchar *path;
	GList *glist;
	GtkCList *clist;
	EDVContext *edv_ctx;
	obj_struct *obj;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	core = CORE(win->core);
	if((clist == NULL) || (core == NULL))
		return;

	edv_ctx = core->edv_ctx;
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	obj = OBJ(gtk_clist_get_row_data(clist, row));
	if(obj == NULL)
		return;

	path = STRDUP(obj->full_path);

	/* Create a new properties dialog */
	edv_window_properties_dialog_new(
		edv_ctx,
		path,
		NULL
	);
	edv_context_flush(edv_ctx);

	g_free(path);
#endif  /* HAVE_LIBENDEAVOUR2 */
}



/*
 *	Set Target Directory.
 */
void WinSetTarDirCB(GtkWidget *widget, gpointer data)
{
	gboolean response;
	gint		npaths = 0,
					nftypes = 0;
	gchar **paths_list = NULL;
	GtkWidget *toplevel;
	fb_type_struct	**ftypes_list = NULL,
					*ftype = NULL;
	core_struct *core;
	win_struct *win = WIN(data);
	if((win == NULL) || FileBrowserIsQuery())
		return;

	if(win->freeze_count > 0)
		return;

	toplevel = win->toplevel;
	core = CORE(win->core);
	if(core == NULL)
		return;

	WinSetBusy(win, TRUE);

	/* Create the file types list */
	FileBrowserTypeListNew(
		&ftypes_list, &nftypes,
		"*.*",
		"All Files"
	);

	/* Query the user to set the target directory */
	FileBrowserSetTransientFor(toplevel);
	response = FileBrowserGetResponse(
		"Set Target Directory",
		"Set", "Cancel",
		WinGetTarDir(win),
		ftypes_list, nftypes,
		&paths_list, &npaths,
		&ftype
	);
	FileBrowserSetTransientFor(NULL);

	/* Got user response? */
	if(response && (npaths > 0))
	{
		gchar *path = STRDUP(paths_list[0]);
		if(path != NULL)
		{
			/* Check if the directory does not exist yet */
			struct stat stat_buf;
			if(stat((const char *)path, &stat_buf))
			{
				const gint error_code = (gint)errno;
				if(error_code == ENOENT)
				{
					/* Query the user to create a new directory */
#ifdef HAVE_LIBENDEAVOUR2
					gint response;
					edv_play_sound_question(core->edv_ctx);
					CDialogSetTransientFor(toplevel);
					response = CDialogGetResponse(
"Create New Directory?",
"The specified directory does not exist,\n\
do you want to create it?",
						NULL,
						CDIALOG_ICON_QUESTION,
						CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
						CDIALOG_BTNFLAG_YES
					);
					CDialogSetTransientFor(NULL);
					if(response == CDIALOG_RESPONSE_YES)
					{
						GList *new_paths_list = NULL;
						if(edv_directory_create(
							path,
							TRUE,
							&new_paths_list
						))
						{
							const gint error_code = (gint)errno;
							gchar *msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
								path,
								g_strerror(error_code)
							);
							edv_play_sound_error(core->edv_ctx);
							CDialogSetTransientFor(toplevel);
							CDialogGetResponse(
								"Create Directory Failed",
								msg,
								NULL,
								CDIALOG_ICON_ERROR,
								CDIALOG_BTNFLAG_OK,
								CDIALOG_BTNFLAG_OK
							);
							g_free(msg);
							CDialogSetTransientFor(NULL);
						}
						else
						{
							GList *glist;
							for(glist = new_paths_list;
								glist != NULL;
								glist = g_list_next(glist)
							)
								edv_notify_queue_vfs_object_added(
									core->edv_ctx,
									(const gchar *)glist->data
								);
							edv_context_flush(core->edv_ctx);
						}
						g_list_foreach(new_paths_list, (GFunc)g_free, NULL);
						g_list_free(new_paths_list);
					}
#else
					gint response;
					CDialogSetTransientFor(toplevel);
					response = CDialogGetResponse(
"Create New Directory?",
"The specified directory does not exist,\n\
do you want to create it?",
						NULL,
						CDIALOG_ICON_QUESTION,
						CDIALOG_BTNFLAG_YES | CDIALOG_BTNFLAG_NO,
						CDIALOG_BTNFLAG_YES
					);
					CDialogSetTransientFor(NULL);
					if(response == CDIALOG_RESPONSE_YES)
					{
						const mode_t m = S_IRWXU | S_IRWXG | S_IRWXO;
						if(rmkdir((char *)path, m))
						{
							const gint error_code = (gint)errno;
							gchar *msg = g_strdup_printf(
"%s:\n\
\n\
    %s",
								path,
								g_strerror(error_code)
							);
							CDialogSetTransientFor(toplevel);
							CDialogGetResponse(
								"Create Directory Failed",
								msg,
								NULL,
								CDIALOG_ICON_ERROR,
								CDIALOG_BTNFLAG_OK,
								CDIALOG_BTNFLAG_OK
							);
							g_free(msg);
							CDialogSetTransientFor(NULL);
						}
					}
#endif
				}
			}
#ifdef S_ISDIR
			else if(!S_ISDIR(stat_buf.st_mode))
#else
			else if(TRUE)
#endif
			{
#ifdef HAVE_LIBENDEAVOUR2
				edv_play_sound_warning(core->edv_ctx);
#endif
				CDialogSetTransientFor(toplevel);
				CDialogGetResponse(
"Not Directory",
"The specified path is not a directory.",
					NULL,
					CDIALOG_ICON_WARNING,
					CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
				);
				CDialogSetTransientFor(NULL);
			}

			/* Set the new target directory */
			WinSetTarDir(
				win,
				path
			);

			g_free(path);
		}
	}

	/* Delete file types list */
	FileBrowserDeleteTypeList(ftypes_list, nftypes);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Convert.
 */
void WinConvertCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinConvert(win);
}

/*
 *	Convert Selected.
 */
void WinConvertSelectedCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinConvertSelected(win);
}

/*
 *	Stop.
 */
void WinStopCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;
			   
	if(win->freeze_count > 0)
		return;

	win->stop_count++;
}

/*
 *	Remove Converted.
 */
void WinRemoveConvertedCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);

	WinSetRemoveConverted(
		win,
		win->remove_converted ? FALSE : TRUE
	);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Recusrive.
 */
void WinRecursiveCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);

	WinSetRecursive(
		win,
		win->recursive ? FALSE : TRUE
	);

	WinUpdate(win);
	WinSetBusy(win, FALSE);
}


/*
 *	Open.
 */
void WinOpenCB(GtkWidget *widget, gpointer data)
{
#if defined(HAVE_LIBENDEAVOUR2)
	gint row;
	GList *glist, *paths_list;
	GtkCList *clist;
	EDVContext *edv_ctx;
	obj_struct *obj;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	core = CORE(win->core);
	if((clist == NULL) || (core == NULL))
		return;

	edv_ctx = core->edv_ctx;
	glist = clist->selection_end;
	row = (glist != NULL) ? (gint)glist->data : -1;
	obj = OBJ(gtk_clist_get_row_data(clist, row));
	if(obj == NULL)
		return;

	paths_list = NULL;
	paths_list = g_list_append(
		paths_list, STRDUP(obj->full_path)
	);

	/* Open the object */
	edv_open(edv_ctx, paths_list, NULL);
	edv_context_flush(edv_ctx);

	g_list_foreach(paths_list, (GFunc)g_free, NULL);
	g_list_free(paths_list);
#endif  /* HAVE_LIBENDEAVOUR2 */
}

/*
 *	Refresh.
 */
void WinRefreshCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	WinSetBusy(win, TRUE);




	WinUpdate(win);
	WinSetBusy(win, FALSE);
}

/*
 *	Preview Source.
 */
void WinPreviewSrcCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	CfgItem *cfg_list;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	w = win->src_preview;
	core = CORE(win->core);
	if((w == NULL) || (core == NULL))
		return;

	WinSetBusy(win, TRUE);

	cfg_list = core->cfg_list;

	if(win->preview_src)
	{
		gtk_widget_hide(w);
		win->preview_src = FALSE;
	}
	else
	{
		gtk_widget_show(w);
		win->preview_src = TRUE;
	}

	WinUpdate(win);

	CFG_SET_B(
		CFG_PARM_WIN_PREVIEW_SRC,
		win->preview_src
	);

	WinSetBusy(win, FALSE);
}

/*
 *	Preview Source Show Details
 */
void WinPreviewSrcShowDetailsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	CfgItem *cfg_list;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	w = win->src_preview;
	core = CORE(win->core);
	if((w == NULL) || (core == NULL))
		return;

	WinSetBusy(win, TRUE);

	cfg_list = core->cfg_list;

	if(win->preview_src_show_details)
	{
		WinPreviewSetShowDetails(w, FALSE);
		win->preview_src_show_details = FALSE;
	}
	else
	{
		WinPreviewSetShowDetails(w, TRUE);
		win->preview_src_show_details = TRUE;
	}

	WinUpdate(win);

	CFG_SET_B(
		CFG_PARM_WIN_PREVIEW_SRC_SHOW_DETAILS,
		win->preview_src_show_details
	);

	WinSetBusy(win, FALSE);
}

/*
 *	Preview Converted.
 */
void WinPreviewTarCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	CfgItem *cfg_list;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	w = win->tar_preview;
	core = CORE(win->core);
	if((w == NULL) || (core == NULL))
		return;

	WinSetBusy(win, TRUE);

	cfg_list = core->cfg_list;

	if(win->preview_tar)
	{
		gtk_widget_hide(w);
		win->preview_tar = FALSE;
	}
	else
	{
		gtk_widget_show(w);
		win->preview_tar = TRUE;
	}

	WinUpdate(win);

	CFG_SET_B(
		CFG_PARM_WIN_PREVIEW_TAR,
		win->preview_tar
	);

	WinSetBusy(win, FALSE);
}

/*
 *	Preview Converted Show Details
 */
void WinPreviewTarShowDetailsCB(GtkWidget *widget, gpointer data)
{
	GtkWidget *w;
	CfgItem *cfg_list;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	w = win->tar_preview;
	core = CORE(win->core);
	if((w == NULL) || (core == NULL))
		return;

	WinSetBusy(win, TRUE);

	cfg_list = core->cfg_list;

	if(win->preview_tar_show_details)
	{
		WinPreviewSetShowDetails(w, FALSE);
		win->preview_tar_show_details = FALSE;
	}
	else
	{
		WinPreviewSetShowDetails(w, TRUE);
		win->preview_tar_show_details = TRUE;
	}

	WinUpdate(win);

	CFG_SET_B(
		CFG_PARM_WIN_PREVIEW_TAR_SHOW_DETAILS,
		win->preview_tar_show_details
	);

	WinSetBusy(win, FALSE);
}

/*
 *	Show/Hide Tool Bar.
 */
void WinViewToolBarCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);

	if(WinToolBarIsShown(win))
		WinToolBarSetShow(win, FALSE);
	else
		WinToolBarSetShow(win, TRUE);

	WinSetBusy(win, FALSE);
}

/*
 *	Show/Hide Status Bar.
 */
void WinViewStatusBarCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	WinSetBusy(win, TRUE);

	if(WinStatusBarIsShown(win))
		WinStatusBarSetShow(win, FALSE);
	else
		WinStatusBarSetShow(win, TRUE);

	WinSetBusy(win, FALSE);
}


/*
 *	Help Contents.
 */
void WinHelpContentsCB(GtkWidget *widget, gpointer data)
{
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	Help(
		CORE(win->core),
		"Contents",
		win->toplevel
	);
}


/*
 *	New ImgWorks.
 */
void WinNewImgWorksCB(GtkWidget *widget, gpointer data)
{
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv_ctx;
#endif
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)             
		return;

	core = CORE(win->core);
	if(core == NULL)
		return;

#ifdef HAVE_LIBENDEAVOUR2
	edv_ctx = core->edv_ctx;
#endif

#if 1
	if(win->toplevel != NULL)
	{
#define PROG_PATH	PREFIX "/bin/imgworks"
		GtkWidget *toplevel = win->toplevel;
		GdkWindow *window = toplevel->window;
#ifdef HAVE_LIBENDEAVOUR2
		const gint	x_offset = edv_get_i(
					edv_ctx,
					EDV_CFG_PARM_WINDOW_CASCADE_OFFSET_X
				),
				y_offset = edv_get_i(
					edv_ctx,
					EDV_CFG_PARM_WINDOW_CASCADE_OFFSET_Y
				);
		gchar *shell_prog;
		const gchar	*shell_cmd = edv_get_s(
					edv_ctx,
					EDV_CFG_PARM_PROG_SHELL
				),
				*shell_args = edv_strarg(
					shell_cmd,
					&shell_prog,
					TRUE,		/* Parse escapes */
					TRUE		/* Parse quotes */
				);
#else
		const gint	x_offset = 30,
				y_offset = 30;
#endif
		GtkAllocation *a = &toplevel->allocation;
		const gint	width = a->width,
				height = a->height;
		gint	x = 0, y = 0,
			pid;
		gchar *cmd;

		/* Calculate the geometry with cascade offset for the new
		 * ImgWorks window
		 */
		if(window != NULL)
			gdk_window_get_root_origin(
				window,
				&x, &y
			);
		x += x_offset;
		y += y_offset;

		/* Format the command to run another ImgWorks */
		cmd = g_strdup_printf(
"%s --geometry %ix%i%s%i%s%i",
			PROG_PATH,
			width, height,
			(x < 0) ? "" : "+", x,
			(y < 0) ? "" : "+", y
		);
#ifdef HAVE_LIBENDEAVOUR2
		/* Execute the command to run another ImgWorks */
		pid = edv_system_shell(
			cmd,
			shell_prog,
			shell_args
		);
		g_free(shell_prog);
#else
		/* Execute the command to run another ImgWorks */
		pid = Exec(cmd);
#endif
		if(pid < 0)
		{
			/* Unable to execute the command */
			gchar *s = g_strdup_printf(
"Unable to execute the command:\n\
\n\
    %s",
				cmd
			);
#ifdef HAVE_LIBENDEAVOUR2
			edv_play_sound_error(edv_ctx);
#endif
			CDialogSetTransientFor(toplevel);
			CDialogGetResponse(  
"Create New ImgWorks Failed",
				s,
				NULL,
				CDIALOG_ICON_ERROR,
				CDIALOG_BTNFLAG_OK,
				CDIALOG_BTNFLAG_OK
			);
			CDialogSetTransientFor(NULL);
			g_free(s);
		}

		g_free(cmd);
#undef PROG_PATH
	}
#else
/* Creating another ImgWorks on the same process is a problem because 
 * the conversion function works poorly with multiple converts
 */
	MainNewImgWorks(core);
#endif
}

/*
 *	New Image Browser.
 */
void WinNewImbrCB(GtkWidget *widget, gpointer data)
{
#if defined(HAVE_LIBENDEAVOUR2)
	gchar *path;
	GtkCList *clist;
	EDVContext *edv_ctx;
	core_struct *core;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	clist = (GtkCList *)win->focus_clist;
	core = CORE(win->core);
	if(core == NULL)
		return;

	edv_ctx = core->edv_ctx;

	/* Get the startup path as the target directory */
	path = STRDUP(WinGetTarDir(win));

	/* If no target directory is specified then get the path of the
	 * last selected object
	 */
	if(STRISEMPTY(path))
	{
		GList *glist = (clist != NULL) ? clist->selection_end : NULL;
		const gint row = (glist != NULL) ? (gint)glist->data : -1;
		obj_struct *obj = (row > -1) ?
			OBJ(gtk_clist_get_row_data(clist, row)) : NULL;
		g_free(path);
		path = (obj != NULL) ? STRDUP(obj->full_path) : NULL;
	}

	/* If the path's destination is not a directory then replace it
	 * with its parent
	 */
	if(!edv_path_is_directory(path))
	{
		gchar *s = g_dirname(path);
		if(s != NULL)
		{
			g_free(path);
			path = s;
		}
	}

	/* Create a new image browser */
	edv_window_image_browser_new(edv_ctx, path);
	edv_context_flush(edv_ctx);

	g_free(path);
#endif  /* HAVE_LIBENDEAVOUR2 */
}


/*
 *	About.
 */
void WinHelpAboutCB(GtkWidget *widget, gpointer data)
{
	gchar *mesg, *s;
	GtkWidget *toplevel;
	win_struct *win = WIN(data);
	if(win == NULL)
		return;

	if(win->freeze_count > 0)
		return;

	win->freeze_count++;
	WinSetBusy(win, TRUE);

	toplevel = win->toplevel;

	/* Format message */
	mesg = STRDUP(
PROG_NAME "\nVersion " PROG_VERSION "\n\n" PROG_COPYRIGHT
	);
#ifdef HAVE_IMAGE_MAGICK
	if(TRUE)
	{
		unsigned long n;
		s = g_strconcat(
			mesg,
			"\nConversion Library: ",
			(const gchar *)MagickGetVersion(&n),
			"\n",
			NULL
		);
		g_free(mesg);
		mesg = s;
	}
#endif

	CDialogSetTransientFor(toplevel);
	CDialogGetResponseIconData(
		"About " PROG_NAME,
		mesg,
		NULL,
		(guint8 **)icon_imgworks_32x32_xpm,
		CDIALOG_BTNFLAG_OK, CDIALOG_BTNFLAG_OK
	);
	g_free(mesg);
	CDialogSetTransientFor(NULL);

	win->freeze_count--;
	WinSetBusy(win, FALSE);
}
